""" Checksum utils
"""

import hashlib

from core_main_app.commons.exceptions import CoreError

CHECKSUM_ALGORITHMS = {
    "MD5": hashlib.md5,
    "SHA1": hashlib.sha1,
    "SHA256": hashlib.sha256,
    "SHA512": hashlib.sha512,
}


def compute_checksum(file, checksum_algorithm, block_size=4096):
    """Compute checksum for a file content (bytes)

    Args:
        file:
        checksum_algorithm:
        block_size:

    Returns:

    """
    # Check if selected algorithm is in the list
    if checksum_algorithm not in CHECKSUM_ALGORITHMS.keys():
        raise CoreError(
            f"CHECKSUM_ALGORITHM needs to be in: {CHECKSUM_ALGORITHMS.keys()}"
        )

    # Init hasher with selected algorithm
    hasher = CHECKSUM_ALGORITHMS[checksum_algorithm]()

    # If param is bytes
    if isinstance(file, bytes):
        # Update hasher
        hasher.update(file)
        # Return hash
        return hasher.hexdigest()

    # If file, read file, chunk by chunk
    for chunk in iter(lambda: file.read(block_size), b""):
        # Hash each chuck
        hasher.update(chunk)
    # Return hash
    return hasher.hexdigest()



"""
Template Version Manager API
"""
from core_main_app.access_control.exceptions import AccessControlError

from core_main_app.access_control.api import is_superuser
from core_main_app.access_control.decorators import access_control
from core_main_app.commons import exceptions
from core_main_app.components.template import api as template_api
from core_main_app.components.template.access_control import (
    can_read_id,
    can_read_global,
)
from core_main_app.components.template_version_manager.access_control import (
    can_write,
    can_write_list,
)
from core_main_app.components.template_version_manager.models import (
    TemplateVersionManager,
)
from core_main_app.components.version_manager import api as version_manager_api
from core_main_app.components.version_manager.access_control import (
    can_read_list,
)
from core_main_app.components.version_manager.utils import (
    get_latest_version_name,
    get_version_name,
)


@access_control(can_read_id)
def get_by_id(version_manager_id, request):
    """Get a template version manager by its id.

    Args:
        version_manager_id:
        request:

    Returns:

    """
    return TemplateVersionManager.get_by_id(version_manager_id)


@access_control(can_write)
def insert(template_version_manager, template, request):
    """Add a version to a template version manager.

    Args:
        template_version_manager:
        template:
        request:

    Returns:

    """
    # save the template in database
    template_api.upsert(template, request=request)
    try:
        # create a version manager
        version_manager_api.upsert(template_version_manager, request=request)
        # set version manager
        template.version_manager = template_version_manager
        # set current version
        if len(template_version_manager.versions) == 0:
            template.is_current = True
        # update template
        template.display_name = get_latest_version_name(
            template_version_manager
        )
        # save template
        template.save_template()
        # return version manager
        return template_version_manager
    except Exception as exception:
        template.delete()
        raise exception


@access_control(can_write)
def edit_title(template_version_manager, title, request):
    """Edit template version manager title.

    Args:
        template_version_manager:
        title:
        request:

    Returns:

    """
    # set version template manager title
    template_version_manager.title = title
    # save template version manager
    template_version_manager.save_template_version_manager()
    # update templates display names
    _update_templates_display_name(template_version_manager, request=request)


def _update_templates_display_name(template_version_manager, request):
    """Update templates display name.

    Args:
        template_version_manager:
        request:

    Returns:

    """
    # Iterate versions
    for i in range(0, len(template_version_manager.versions)):
        # get template id from list of versions
        template_id = template_version_manager.versions[i]
        # get template from template id
        template = template_api.get_by_id(template_id, request=request)
        # get display name for the template
        display_name = get_version_name(template_version_manager.title, i + 1)
        # update template's display name
        template_api.set_display_name(template, display_name, request=request)


@access_control(can_read_global)
def get_global_version_managers(request, _cls=True):
    """Get all global version managers of a template.

    Args:
        request:
        _cls:

    Returns:

    """
    return TemplateVersionManager.get_global_version_managers(_cls)


@access_control(can_read_global)
def get_active_global_version_manager(request, _cls=True):
    """Return all active Version Managers with user set to None.

    Args:
        request:
        _cls:

    Returns:

    """
    return TemplateVersionManager.get_active_global_version_manager(_cls)


# NOTE: access control, filter by user in request
def get_active_version_manager_by_user_id(request, _cls=True):
    """Return all active Version Managers with given user id.

    Returns:

    """
    return TemplateVersionManager.get_active_version_manager_by_user_id(
        str(request.user.id), _cls
    )


# NOTE: access control, filter by user in request
def get_all_by_user_id(request, _cls=True):
    """Return all Template Version Managers with given user id.

    Returns:

    """
    return TemplateVersionManager.get_all_version_manager_by_user_id(
        str(request.user.id), _cls
    )


@access_control(is_superuser)
def get_all(request, _cls=True):
    """Return all Template Version Managers.

    Returns:

    """
    return TemplateVersionManager.get_all_version_manager(_cls)


@access_control(can_read_list)
def get_by_id_list(list_id, request):
    """Get a version managers with the given id list.

    Args:
        list_id:
        request

    Returns:

    """
    return TemplateVersionManager.get_by_id_list(list_id)


@access_control(can_read_global)
def get_active_global_version_manager_by_title(version_manager_title, request):
    """Return active Version Manager by its title with user set to None.

    Args:
        version_manager_title: Version Manager title
        request:

    Returns:
        Version Manager instance

    """
    return TemplateVersionManager.get_active_global_version_manager_by_title(
        version_manager_title
    )


def _update_template_version_manager_ordering(list_id, request):
    """Update template version manager ordering.

    Args:
        list_id:
        request:

    Returns:

    """

    tvm_list = dict(
        [(tvm.id, tvm) for tvm in get_by_id_list(list_id, request)]
    )
    # Check if all templates have been found
    if len(list_id) > len(tvm_list):
        raise exceptions.DoesNotExist(
            "One or more templates could not be found"
        )
    template_version_manager_sorted_by_id_list = [
        tvm_list[int(id)] for id in list_id
    ]

    for counter, template_version_manager in enumerate(
        template_version_manager_sorted_by_id_list, 1
    ):
        template_version_manager.display_rank = counter
        template_version_manager.save()


@access_control(can_write_list)
def update_user_template_ordering(list_id, request):
    """Update templates ordering.

    Args:
        list_id:
        request:

    Returns:

    """
    if any(
        tvm.user is None or tvm.user != str(request.user.id)
        for tvm in get_by_id_list(list_id, request)
    ):
        raise AccessControlError(
            "You don't have the rights to perform this action."
        )
    _update_template_version_manager_ordering(list_id, request)


@access_control(is_superuser)
def update_global_template_ordering(list_id, request):
    """Update templates ordering.

    Args:
        list_id:
        request:

    Returns:

    """
    if any(tvm.user is not None for tvm in get_by_id_list(list_id, request)):
        raise AccessControlError(
            "You don't have the rights to perform this action."
        )
    _update_template_version_manager_ordering(list_id, request)



""" Web page model
"""
from django.core.exceptions import ObjectDoesNotExist
from django.db import models

from core_main_app.commons import exceptions
from core_main_app.commons.enums import WEB_PAGE_TYPES


class WebPage(models.Model):
    """Represents a WebPage"""

    type = models.IntegerField()
    content = models.TextField()

    @staticmethod
    def get_by_type(page_type):
        """Get a WebPage given its type

        Parameters:
            page_type (str): page type

        Returns:
            Web Page corresponding to the given type
        """
        try:
            return WebPage.objects.get(type=WEB_PAGE_TYPES[page_type])
        except ObjectDoesNotExist as e:
            raise exceptions.DoesNotExist(str(e))
        except Exception as ex:
            raise exceptions.ModelError(str(ex))

    @staticmethod
    def delete_by_type(page_type_key):
        """Delete all WebPage with the given type key

        Args:
            page_type_key (int): page type key

        Returns:

        """
        return WebPage.objects.filter(type=page_type_key).delete()



"""Utils for Django raw query
"""

from django.conf import settings

from core_main_app.utils.raw_query.common import (
    check_user_filter,
    check_workspace_filter,
)


def add_access_criteria(
    query, accessible_workspaces, user, workspace_filter=None, user_filter=None
):
    """Add access criteria to the query.

    Args:
        query:
        accessible_workspaces:
        user:
        workspace_filter:
        user_filter:

    Returns:

    """
    # get access criteria to update query
    access_criteria = _get_accessible_criteria(
        accessible_workspaces, user, workspace_filter, user_filter
    )
    # add access criteria to original query
    query &= access_criteria
    # return query
    return query


def _get_accessible_criteria(
    accessible_workspaces, user, workspace_filter=None, user_filter=None
):
    """Get accessible criteria.

    Args:
        accessible_workspaces:
        user:
        workspace_filter:
        user_filter:

    Returns:

    """
    if settings.MONGODB_INDEXING:
        from mongoengine.queryset.visitor import Q

        workspace_key = "_workspace_id"
    else:
        from django.db.models import Q

        workspace_key = "workspace"
    # if superuser don't check permissions, only apply filters
    if user.is_superuser:
        # create query
        query_filter = Q()
        # if workspaces provided
        if workspace_filter:
            # add filter by workspace
            query_filter &= get_workspace_query(workspace_filter)
        # if user ids provided
        if user_filter:
            # add filter by users
            query_filter &= Q(user_id__in=user_filter)
        # return the query
        return query_filter

    # check if user can filter by user
    check_user_filter(user_filter, user)
    # check if user can filter by workspace
    check_workspace_filter(workspace_filter, accessible_workspaces)

    # build workspace query: all accessible workspaces or only selected workspaces among accessible ones
    workspace_query = (
        [
            workspace
            for workspace in workspace_filter
            if workspace in accessible_workspaces
        ]
        if workspace_filter
        else accessible_workspaces
    )
    # workspace should be in list of accessible workspaces
    access_criteria = Q(**{f"{workspace_key}__in": workspace_query})
    # if only workspace filter set, return access criteria
    if workspace_filter and not user_filter:
        return access_criteria
    # if connected user making the query
    if user and user.id:
        # if user filter is set
        if user_filter:
            # filter on user id
            access_criteria &= Q(user_id=str(user.id))
        else:
            # otherwise include user's own data to results
            access_criteria |= Q(user_id=str(user.id))
    # return access criteria
    return access_criteria


def get_workspace_query(list_workspace):
    """Return Q object to match list of workspaces

    Args:
        list_workspace:

    Returns:

    """
    if settings.MONGODB_INDEXING:
        from mongoengine.queryset.visitor import Q

        workspace_key = "_workspace_id"
    else:
        from django.db.models import Q

        workspace_key = "workspace"

    # create initial query that returns all workspaces
    if len(list_workspace) == 0:
        return Q(**{f"{workspace_key}__in": []})
    # create empty query
    workspace_q_list = Q()
    for workspace_id in list_workspace:
        # if workspace id is None
        if workspace_id is None:
            # add query: OR workspace is null
            workspace_q_list |= Q(**{f"{workspace_key}__exact": None})
        else:  # workspace is not null
            # add query: OR workspace is equal to id
            workspace_q_list |= Q(**{f"{workspace_key}__exact": workspace_id})
    # return query
    return workspace_q_list



""" Access Control Test for User Preferences
"""
from unittest.mock import patch, Mock

from django.contrib.auth.models import AnonymousUser
from django.core.exceptions import ValidationError
from tests.components.user_preferences.fixtures.fixtures import (
    UserPreferencesFixtures,
)

from core_main_app.access_control.exceptions import AccessControlError
from core_main_app.commons import exceptions
from core_main_app.components.user_preferences import (
    api as user_preferences_api,
)
from core_main_app.components.user_preferences.models import UserPreferences
from core_main_app.utils.integration_tests.integration_base_transaction_test_case import (
    IntegrationTransactionTestCase,
)
from core_main_app.utils.tests_tools.MockUser import create_mock_user

fixture_user_preferences = UserPreferencesFixtures()


class TestUserPreferencesUpsert(IntegrationTransactionTestCase):
    """Test User Preferences Upsert"""

    fixture = fixture_user_preferences

    def test_upsert_user_preferences_as_owner_creates_preferences(self):
        """test_upsert_user_preferences_as_owner_creates_preferences

        Args:

        Returns:

        """
        # Arrange
        user = create_mock_user("4")

        user_preferences = UserPreferences(user_id="4", timezone="GMT")
        # Act
        result = user_preferences_api.upsert(user_preferences, user)

        # Assert
        self.assertEquals(result.timezone, user_preferences.timezone)
        self.assertEquals(result.user_id, user.id)

    def test_upsert_user_preferences_as_superuser_creates_preferences(self):
        """test_upsert_user_preferences_as_superuser_creates_preferences

        Args:

        Returns:

        """
        # Arrange
        user = create_mock_user("1", is_superuser=True)

        user_preferences = UserPreferences(user_id="4", timezone="GMT")
        # Act
        result = user_preferences_api.upsert(user_preferences, user)

        # Assert
        self.assertEquals(result.timezone, user_preferences.timezone)
        self.assertEquals(result.user_id, user_preferences.user_id)

    def test_upsert_user_preferences_as_user_raises_access_control(self):
        """test_upsert_user_preferences_as_user_raises_access_control

        Args:

        Returns:

        """
        # Arrange
        user = create_mock_user("1")

        user_preferences = UserPreferences(user_id="4", timezone="GMT")
        # Act
        with self.assertRaises(AccessControlError):
            user_preferences_api.upsert(user_preferences, user)

    def test_upsert_user_preferences_as_anonymous_raises_access_control(self):
        """test_upsert_user_preferences_as_anonymous_raises_access_control

        Args:

        Returns:

        """
        # Arrange
        user = AnonymousUser()

        user_preferences = UserPreferences(user_id="4", timezone="GMT")
        # Act
        with self.assertRaises(AccessControlError):
            user_preferences_api.upsert(user_preferences, user)

    def test_upsert_user_preferences_as_owner_updates_preferences(self):
        """test_upsert_user_preferences_as_owner_updates_preferences

        Args:

        Returns:

        """
        # Arrange
        user = create_mock_user("1")

        user_preferences = self.fixture.user_preferences_1
        user_preferences.timezone = "Europe/Paris"
        # Act
        result = user_preferences_api.upsert(user_preferences, user)

        # Assert
        self.assertEquals(result.timezone, "Europe/Paris")
        self.assertEquals(result.user_id, user.id)

    def test_upsert_others_user_preferences_as_superuser_updates_preferences(
        self,
    ):
        """test_upsert_others_user_preferences_as_superuser_updates_preferences

        Args:

        Returns:

        """
        # Arrange
        user = create_mock_user("1", is_superuser=True)

        user_preferences = self.fixture.user_preferences_3
        user_preferences.timezone = "Europe/Paris"
        # Act
        result = user_preferences_api.upsert(user_preferences, user)

        # Assert
        self.assertEquals(result.timezone, "Europe/Paris")
        self.assertEquals(result.user_id, user_preferences.user_id)

    def test_upsert_others_user_preferences_as_user_raises_access_control(
        self,
    ):
        """test_upsert_others_user_preferences_as_user_raises_access_control

        Args:

        Returns:

        """
        # Arrange
        user = create_mock_user("1")

        user_preferences = self.fixture.user_preferences_2
        user_preferences.timezone = "fake_zone"

        # Act
        with self.assertRaises(AccessControlError):
            user_preferences_api.upsert(user_preferences, user)

    def test_upsert_others_user_preferences_as_anonymous_raises_access_control(
        self,
    ):
        """test_upsert_others_user_preferences_as_anonymous_raises_access_control

        Args:

        Returns:

        """
        # Arrange
        user = AnonymousUser()

        user_preferences = self.fixture.user_preferences_2
        user_preferences.timezone = "fake_zone"

        # Act
        with self.assertRaises(AccessControlError):
            user_preferences_api.upsert(user_preferences, user)

    def test_upsert_user_preferences_as_owner_raises_error_when_user_had_already_preferences(
        self,
    ):
        """test_upsert_user_preferences_as_owner_raises_error_when_user_had_already_preferences

        Args:

        Returns:

        """
        # Arrange
        user = create_mock_user("1")

        user_preferences = UserPreferences(user_id="1", timezone="GMT")

        # Act  # Assert
        with self.assertRaises(ValidationError):
            user_preferences_api.upsert(user_preferences, user)

    def test_upsert_others_user_preferences_as_superuser_raises_error_when_user_had_already_preferences(
        self,
    ):
        """test_upsert_others_user_preferences_as_superuser_raises_error_when_user_had_already_preferences

        Args:

        Returns:

        """
        # Arrange
        user = create_mock_user("0", is_superuser=True)

        user_preferences = UserPreferences(user_id="1", timezone="GMT")

        # Act  # Assert
        with self.assertRaises(ValidationError):
            user_preferences_api.upsert(user_preferences, user)


class TestUserPreferencesDelete(IntegrationTransactionTestCase):
    """Test User Preferences Delete"""

    fixture = fixture_user_preferences

    def test_delete_user_preferences_as_owner_delete_preferences(self):
        """test_delete_user_preferences_as_owner_delete_preferences

        Args:

        Returns:

        """
        # Arrange
        user = create_mock_user("1")

        # Act
        user_preferences_api.delete(self.fixture.user_preferences_1, user)

        self.assertEquals(self.fixture.user_preferences_1.id, None)

    def test_delete_user_preferences_as_superuser_delete_preferences(self):
        """test_delete_user_preferences_as_superuser_delete_preferences

        Args:

        Returns:

        """
        # Arrange
        user = create_mock_user("1")

        # Act
        user_preferences_api.delete(self.fixture.user_preferences_1, user)

        self.assertEquals(self.fixture.user_preferences_1.id, None)

    def test_delete_others_user_preferences_as_superuser_delete_preferences(
        self,
    ):
        """test_delete_others_user_preferences_as_superuser_delete_preferences

        Args:

        Returns:

        """
        # Arrange
        user = create_mock_user("1", is_superuser=True)

        # Act
        user_preferences_api.delete(self.fixture.user_preferences_3, user)

        self.assertEquals(self.fixture.user_preferences_3.id, None)

    def test_delete_others_user_preferences_as_user_raises_access_control(
        self,
    ):
        """test_delete_others_user_preferences_as_user_raises_access_control

        Args:

        Returns:

        """
        # Arrange
        user = create_mock_user("1")

        user_preferences = self.fixture.user_preferences_2
        user_preferences.timezone = "fake_zone"

        # Act
        with self.assertRaises(AccessControlError):
            user_preferences_api.delete(user_preferences, user)

    def test_update_others_user_preferences_as_anonymous_raises_access_control(
        self,
    ):
        """test_update_others_user_preferences_as_anonymous_raises_access_control

        Args:

        Returns:

        """
        # Arrange
        user = AnonymousUser()

        user_preferences = self.fixture.user_preferences_2
        user_preferences.timezone = "fake_zone"

        # Act
        with self.assertRaises(AccessControlError):
            user_preferences_api.delete(user_preferences, user)


class TestUserPreferencesGetByUser(IntegrationTransactionTestCase):
    """Test User Preferences Get By User"""

    fixture = fixture_user_preferences

    def test_get_user_preferences_by_user_as_owner_returns_preferences(self):
        """test_get_user_preferences_by_user_as_owner_returns_preferences

        Args:

        Returns:

        """
        # Arrange
        user = create_mock_user("1")

        # Act
        result = user_preferences_api.get_by_user(user)

        # Assert
        self.assertEquals(result.timezone, "UTC")
        self.assertEquals(result.user_id, user.id)

    @patch(
        "core_main_app.components.user_preferences.models.UserPreferences.get_by_user"
    )
    def test_get_user_preferences_with_wrong_owner_raises_acl_error(
        self, mock_get_by_user
    ):
        """test_get_user_preferences_with_wrong_owner_raises_acl_error

        Args:

        Returns:

        """
        # Arrange
        user = create_mock_user("1")
        mock_user_preferences = Mock(UserPreferences)
        mock_user_preferences.user_id = "2"
        mock_get_by_user.return_value = mock_user_preferences

        # Act + Assert
        with self.assertRaises(AccessControlError):
            user_preferences_api.get_by_user(user)

    def test_get_others_user_preferences_with_no_user_raises_access_control_error(
        self,
    ):
        """test_get_others_user_preferences_by_user_as_user_raises_access_control_error

        Args:

        Returns:

        """
        # Act  # Assert
        with self.assertRaises(AccessControlError):
            user_preferences_api.get_by_user(None)

    def test_get_others_user_preferences_by_user_as_anonymous_raises_access_control_error(
        self,
    ):
        """test_get_others_user_preferences_by_user_as_anonymous_raises_access_control_error

        Args:

        Returns:

        """
        # Arrange
        user = AnonymousUser()

        # Act  # Assert
        with self.assertRaises(AccessControlError):
            user_preferences_api.get_by_user(user)

    def test_get_user_preferences_by_user_as_user_raises_dne_exception_when_no_preferences_founded(
        self,
    ):
        """test_get_user_preferences_by_user_as_user_raises_dne_exception_when_no_preferences_founded

        Args:

        Returns:

        """

        # Act + Assert
        with self.assertRaises(exceptions.DoesNotExist):
            user_preferences_api.get_by_user(create_mock_user("0"))

    def test_get_user_preferences_by_user_as_superuser_raises_dne_exception_when_no_preferences_found(
        self,
    ):
        """test_get_user_preferences_by_user_as_superuser_raises_dne_exception_when_no_preferences_found

        Args:

        Returns:

        """

        # Act + Assert
        with self.assertRaises(exceptions.DoesNotExist):
            user_preferences_api.get_by_user(
                create_mock_user("0", is_superuser=True)
            )



""" Unit test Blob
"""

from unittest.case import TestCase
from unittest.mock import patch

from django.contrib.auth.models import User
from tests.components.data.tests_unit import (
    _create_blob,
    _create_data,
    _get_template,
)

from core_main_app.components.blob import api as blob_api
from core_main_app.utils.tests_tools.MockUser import create_mock_user


class TestBlobGetNone(TestCase):
    """Test Blob Get None"""

    def test_blob_get_none_returns_empty_list(self):
        """test_blob_get_none_returns_empty_list

        Args:

        Returns:

        """
        # Act
        result = blob_api.get_none()

        # Assert
        self.assertEqual(len(result), 0)


class TestBlobMetadata(TestCase):
    """Test Blob Metadata"""

    def test_blob_metadata_returns_list_of_data(self):
        """test_blob_metadata_returns_list_of_data

        Returns:

        """
        # Arrange
        blob = _create_blob(user_id="2")
        blob.save()
        template = _get_template()
        template.save()
        data = _create_data(
            data_id=1,
            template=template,
            user_id="2",
            title="new_title",
            content="<tag></tag>",
            blob=blob,
        )
        data.save()
        blob._metadata.set([data])
        user = create_mock_user("2", is_superuser=True)
        # Act
        metadata = blob.metadata(user=user)
        # Assert
        self.assertEqual(list(metadata), [data])

    def test_blob_metadata_returns_empty_list_if_not_set(self):
        """test_blob_metadata_returns_none_if_not_set

        Returns:

        """
        # Arrange
        blob = _create_blob(user_id="2")
        blob.save()
        user = create_mock_user("2", is_superuser=True)
        # Act
        metadata = blob.metadata(user=user)
        # Assert
        self.assertEqual(list(metadata), [])


class TestBlobOwnerName(TestCase):
    """Test Blob Owner Name"""

    @patch.object(User, "objects")
    def test_blob_owner_name(self, user_objects):
        """test_blob_owner_name

        Returns:

        """
        # Arrange
        user = create_mock_user(user_id="2")
        user.username = "user"
        user_objects.get.return_value = user

        blob = _create_blob(user_id="2")

        # Assert
        self.assertEqual(blob.owner_name, user.username)



""" Custom admin site for the Template Xsl Rendering model
"""
from django.contrib import admin


class CustomTemplateXslRenderingAdmin(admin.ModelAdmin):
    """CustomTemplateXslRenderingAdmin"""

    def has_add_permission(self, request, obj=None):
        """Prevent from manually adding Template Xsl rendering"""
        return False



"""Template tag to print date
"""
import pytz
from django import template

register = template.Library()


@register.filter
def print_datetime_utc_unaware(datetime):
    """Print datetime utc.

    Args:
        datetime: datetime timezone unaware

    Returns:

    """
    return datetime.replace(tzinfo=pytz.UTC)



""" Custom admin site for the Data model
"""
from django.conf import settings
from django.contrib import messages
from django.contrib.admin.helpers import ActionForm
from django.forms import ChoiceField

from core_main_app.commons.exceptions import DoesNotExist
from core_main_app.components.user import api as user_api
from core_main_app.components.workspace import api as workspace_api
from core_main_app.utils.admin_site.model_admin_class import (
    get_base_model_admin_class,
)


class UpdateActionForm(ActionForm):
    """Action form for data update"""

    user_id = ChoiceField(label="Owner:", required=False)
    workspace = ChoiceField(label="Workspace:", required=False)

    WORKSPACES_OPTIONS = []
    USER_OPTIONS = []

    def __init__(self, *args, **kwargs):
        # Clear workspace choices
        self.WORKSPACES_OPTIONS = list()
        # Add empty option (no workspace change)
        self.WORKSPACES_OPTIONS.append(("", ""))
        # Add None option (remove from workspace)
        self.WORKSPACES_OPTIONS.append(("None", "Remove from Workspace"))
        # Get all workspaces
        all_workspaces = workspace_api.get_all()
        # Add all workspaces options
        for workspace in all_workspaces:
            self.WORKSPACES_OPTIONS.append((workspace.id, workspace.title))

        # Clear user choices
        self.USERS_OPTIONS = list()
        # Add empty option (no user change)
        self.USERS_OPTIONS.append(("", ""))
        # Get all active users
        all_users = sorted(
            user_api.get_active_users(), key=lambda s: s.username.lower()
        )

        # Add all users options
        for user in all_users:
            self.USERS_OPTIONS.append((user.id, user.username))

        # Init action form
        super().__init__(*args, **kwargs)
        # Set choices
        self.fields["user_id"].choices = list()
        self.fields["user_id"].choices = self.USERS_OPTIONS
        self.fields["workspace"].choices = list()
        self.fields["workspace"].choices = self.WORKSPACES_OPTIONS


def update_data_list(model_admin, request, queryset):
    """Update data list

    Args:
        model_admin:
        request:
        queryset:

    Returns:

    """
    try:
        # Check if user_id parameter provided
        if "user_id" in request.POST and request.POST["user_id"] != "":
            # Get user id
            user_id = (
                None
                if request.POST["user_id"] == "None"
                else str(request.POST["user_id"])
            )
            # Check if user exists
            try:
                user_api.get_user_by_id(user_id)
            except DoesNotExist:
                model_admin.message_user(
                    request, "No user found with this id.", messages.WARNING
                )
            # Update user
            queryset.update(user_id=user_id)
            # No signals on queryset, start update in mongo
            if settings.MONGODB_INDEXING:
                from core_main_app.components.mongo.models import MongoData

                MongoData.update_user_id_from_queryset(queryset, user_id)
            # Display success message
            model_admin.message_user(
                request,
                f"Successfully updated owner of {queryset.count()} Data",
                messages.SUCCESS,
            )
        # Check if workspace_id parameter provided
        if "workspace" in request.POST and request.POST["workspace"] != "":
            # Get workspace
            workspace = (
                None
                if request.POST["workspace"] == "None"
                else request.POST["workspace"]
            )
            # Check if workspace exists
            try:
                workspace = (
                    workspace_api.get_by_id(workspace)
                    if workspace is not None
                    else None
                )
            except DoesNotExist:
                model_admin.message_user(
                    request,
                    "No workspace found with this id.",
                    messages.WARNING,
                )
            # Update workspace
            queryset.update(workspace=workspace)
            if settings.MONGODB_INDEXING:
                from core_main_app.components.mongo.models import MongoData

                workspace_id = workspace.id if workspace else None
                MongoData.update_workspace_id_from_queryset(
                    queryset, workspace_id
                )
            # Display success message
            model_admin.message_user(
                request,
                f"Successfully updated workspace of {queryset.count()} Data",
                messages.SUCCESS,
            )
    except DoesNotExist as ex:
        model_admin.message_user(request, str(ex), messages.ERROR)
    except Exception as ex:
        model_admin.message_user(request, str(ex), messages.ERROR)


class CustomDataAdmin(get_base_model_admin_class("Data")):
    """Custom Data Admin"""

    search_fields = ["title", "vector_column"]
    list_filter = ["template", "user_id", "workspace"]
    list_display = [
        "title",
        "last_modification_date",
        "owner_name",
        "workspace",
    ]
    action_form = UpdateActionForm
    actions = [update_data_list]
    readonly_fields = ["checksum", "xml_file"]
    exclude = ["vector_column", "dict_content"]

    def has_add_permission(self, request, obj=None):
        """Prevent from manually adding data"""
        return False



""" User preferences Access Control
"""
from django.contrib.auth.models import AnonymousUser, User

from core_main_app.access_control.exceptions import AccessControlError
from core_main_app.components.user_preferences.models import UserPreferences


def can_write(func, *args, **kwargs):
    """Can write user preferences.

    Args:
        func:
        args:
        kwargs:

    Returns:

    """
    user = next(
        (arg for arg in args if isinstance(arg, (User, AnonymousUser))),
        None,
    )

    # anonymous cannot write
    if not user or user.is_anonymous:
        raise AccessControlError(
            "The user doesn't have enough rights to edit these preferences."
        )

    # super user
    if user.is_superuser:
        return func(*args, **kwargs)

    document = next(
        (arg for arg in args if isinstance(arg, UserPreferences)), None
    )
    if document.user_id != str(user.id):
        raise AccessControlError(
            "The user doesn't have enough rights to edit these preferences."
        )

    return func(*args, **kwargs)


def can_read(func, user):
    """Can read user preferences.

    Args:
        func:
        user:

    Returns:

    """
    # If anon, raise error
    if user is None or isinstance(user, AnonymousUser):
        raise AccessControlError(
            "The user doesn't have enough rights to access these preferences."
        )

    # If user is superuser return func
    if user.is_superuser:
        return func(user)

    # Get user preferences
    user_preferences = func(user)

    # If user preferences don't belong to user
    if user_preferences.user_id != str(user.id):
        raise AccessControlError(
            "The user doesn't have enough rights to access these preferences."
        )

    # Return preferences
    return user_preferences



""" Raw queries common utils
"""

from core_main_app.access_control.exceptions import AccessControlError


def check_user_filter(user_filter, user):
    """Check that user is allowed to filter by user

    Args:
        user_filter:
        user:

    Returns:

    """

    def _check_user(user_id_filter, user_id):
        """Raise ACL error if user_id_filter different from user id doing request

        Args:
            user_id_filter:
            user_id:

        Returns:

        """
        if user_id_filter and str(user_id_filter) != str(user_id):
            # raise access control error
            raise AccessControlError(
                "The user does not have enough rights to filter by user."
            )

    # if regular user does a query on other user id, raise access control error
    if isinstance(user_filter, list):
        for user_filter_id in user_filter:
            _check_user(user_filter_id, user.id)
    else:
        _check_user(user_filter, user.id)


def check_workspace_filter(workspace_filter, accessible_workspaces):
    """Check that user is allowed to filter by list of workspaces

    Args:
        workspace_filter:
        accessible_workspaces:

    Returns:

    """
    # iterate list of workspace filter provided
    if workspace_filter:
        for workspace in workspace_filter:
            # if workspace is not accessible
            if workspace not in accessible_workspaces:
                # raise access control error
                raise AccessControlError(
                    "The user does not have enough right to filter by workspace."
                )



""" Core Exceptions
"""


class BaseCoreException(Exception):
    """BaseCoreException"""

    def __init__(self, message):
        """Initialize exception

        Args:
            message:
        """
        super().__init__()
        self.message = message

    def __str__(self):
        """Exception as string

        Returns:

        """
        return self.message


class CoreError(BaseCoreException):
    """Exception raised by the Core."""

    def __init__(self, message):
        """Initialize exception

        Args:
            message:
        """
        super().__init__(message)


class ApiError(BaseCoreException):
    """Exception raised by the API."""

    def __init__(self, message):
        """Initialize exception

        Args:
            message:
        """
        super().__init__(message)


class ModelError(BaseCoreException):
    """Generic exception for the model."""

    def __init__(self, message):
        """Initialize exception

        Args:
            message:
        """
        super().__init__(message)


class RestApiError(BaseCoreException):
    """Exception raised by the the REST API."""

    def __init__(self, message):
        """Initialize exception

        Args:
            message:
        """
        super().__init__(message)


class DoesNotExist(BaseCoreException):
    """Exception raised when an object does not exist."""

    def __init__(self, message):
        """Initialize exception

        Args:
            message:
        """
        super().__init__(message)


class NotUniqueError(BaseCoreException):
    """Exception raised when an object is not unique."""

    def __init__(self, message):
        """Initialize exception

        Args:
            message:
        """
        super().__init__(message)


class XMLError(BaseCoreException):
    """Exception raised by XML validation."""

    def __init__(self, message):
        """Initialize exception

        Args:
            message:
        """
        super().__init__(message)


class XSDError(BaseCoreException):
    """Exception raised by XSD validation."""

    def __init__(self, message):
        """Initialize exception

        Args:
            message:
        """
        super().__init__(message)


class LockError(BaseCoreException):
    """Exception raised when an object is locked."""

    def __init__(self, message):
        """Initialize exception

        Args:
            message:
        """
        super().__init__(message)


class PaginationError(BaseCoreException):
    """Exception raised when an error occurs during pagination."""

    def __init__(self, message):
        """Initialize exception

        Args:
            message:
        """
        super().__init__(message)


class BlobDownloaderUrlParseError(BaseCoreException):
    """Exception raised when an error occurs during url parse."""

    def __init__(self, message):
        """Initialize exception

        Args:
            message:
        """
        super().__init__(message)


class BlobDownloaderError(BaseCoreException):
    """Exception raised when an error occurs during blob download."""

    def __init__(self, message):
        """Initialize exception

        Args:
            message:
        """
        super().__init__(message)


class SSLError(BaseCoreException):
    """Exception raised when an error occurs during SSL configuration."""

    def __init__(self, message):
        """Initialize exception

        Args:
            message:
        """
        super().__init__(message)


class QueryError(BaseCoreException):
    """Exception raised when an error occurs regarding queries."""

    def __init__(self, message):
        """Initialize exception

        Args:
            message:
        """
        super().__init__(message)


class DocumentEditingSizeError(BaseCoreException):
    """Exception raised when an edited document is too large."""

    def __init__(self, message):
        """Initialize exception

        Args:
            message:
        """
        super().__init__(message)



""" Serializers for Django User
"""
from django.contrib.auth.models import User
from rest_framework import serializers


class UserSerializer(serializers.ModelSerializer):
    """User serializer"""

    class Meta:
        """Meta"""

        model = User
        fields = ["id", "username", "email", "password"]
        extra_kwargs = {"password": {"write_only": True}}

    def create(self, validated_data):
        """create

        Args:
            validated_data

        Returns:
        """
        password = validated_data.pop("password")
        user = User(**validated_data)
        user.set_password(password)
        user.save()

        return user

    def update(self, instance, validated_data):
        """update

        Args:
            instance:
            validated_data:

        Returns:
        """
        instance.username = validated_data.get("username", instance.username)
        instance.email = validated_data.get("email", instance.email)

        password = validated_data.get("password", None)
        if password is not None:
            instance.set_password(password)

        instance.save()
        return instance



"""Url router for the REST API
"""
from django.urls import re_path

from core_main_app.rest.blob import views as blob_views
from core_main_app.rest.data import views as data_views
from core_main_app.rest.template import views as template_views

urlpatterns = [
    re_path(
        r"^template/(?P<pk>\w+)/download/$",
        template_views.TemplateDownload.as_view(),
        name="core_main_app_rest_template_download",
    ),
    re_path(
        r"^data/download/(?P<pk>\w+)/$",
        data_views.DataDownload.as_view(),
        name="core_main_app_rest_data_download",
    ),
    re_path(
        r"^blob/download/(?P<pk>\w+)/$",
        blob_views.BlobDownload.as_view(),
        name="core_main_app_rest_blob_download",
    ),
]



""" Templatetags to parse date string and convert to datetime object
"""
from datetime import datetime

from django import template
from django.utils import dateparse

register = template.Library()


@register.filter(name="parse_date")
def parse_date(value):
    """Parse date string

    Args:
        value: date string

    Returns: datetime
    """
    if isinstance(value, datetime):
        return value
    try:
        return dateparse.parse_datetime(value)
    except ValueError:
        return None



""" Validate filename test case
"""
from unittest import TestCase

from django.core.exceptions import ValidationError

from core_main_app.utils.validation.regex_validation import validate_filename


class TestValidateFilename(TestCase):
    """Test validate_filename"""

    def test_validate_filename_ok_if_filename_ok(self):
        """test_validate_filename_ok_if_filename_ok

        Returns:

        """
        validate_filename("test.png")

    def test_validate_filename_ok_if_filename_with_underscore_ok(self):
        """test_validate_filename_ok_if_filename_with_underscore_ok

        Returns:

        """
        validate_filename("test_1.png")

    def test_validate_filename_ok_if_filename_with_hyphen_ok(self):
        """test_validate_filename_ok_if_filename_with_hyphen_ok

        Returns:

        """
        validate_filename("test-1.png")

    def test_validate_filename_ok_if_filename_with_all_characters_ok(self):
        """test_validate_filename_ok_if_filename_with_all_characters_ok

        Returns:

        """
        validate_filename("_Test-1.png")

    def test_validate_filename_if_extension_missing(self):
        """test_validate_filename_if_extension_missing

        Returns:

        """
        with self.assertRaises(ValidationError):
            validate_filename("test")



""" Unit tests for migrations utilities
"""
from django.db import OperationalError
from unittest.mock import patch

from django.test import TestCase

from core_main_app.utils.migrations import ensure_migration_applied
from tests.mocks import MockQuerySet


class TestEnsureMigrationApplied(TestCase):
    """Unit tests for ensure_migration_applied function"""

    @staticmethod
    def patch_migration_object(expected_count):
        mock_migration_qs = MockQuerySet()
        mock_migration_qs.item_list = [None] * expected_count

        return mock_migration_qs

    @patch(
        "django.db.migrations.recorder.MigrationRecorder.Migration.objects.filter"
    )
    def test_migration_applied_exits_succesfully(self, mock_migration):
        mock_migration.return_value = self.patch_migration_object(1)

        self.assertIsNone(
            ensure_migration_applied("mock_app", "mock_migration")
        )

    @patch(
        "django.db.migrations.recorder.MigrationRecorder.Migration.objects.filter"
    )
    def test_migration_not_applied_raises_runtime_error(self, mock_migration):
        mock_migration.return_value = self.patch_migration_object(0)

        with self.assertRaises(RuntimeError):
            ensure_migration_applied("mock_app", "mock_migration")

    @patch(
        "django.db.migrations.recorder.MigrationRecorder.Migration.objects.filter"
    )
    def test_operational_error_is_discarded(self, mock_migration):
        mock_migration.side_effect = OperationalError(
            "mock_migration_operational_error"
        )

        self.assertIsNone(
            ensure_migration_applied("mock_app", "mock_migration")
        )



""" Set of functions to define the rules for access control
"""
import logging

from core_main_app.access_control.api import (
    has_perm_publish,
    check_can_read_list,
    can_write_in_workspace,
    check_anonymous_access,
    check_can_read_document,
)
from core_main_app.components.workspace import api as workspace_api
from core_main_app.permissions import rights as rights
from core_main_app.settings import (
    VERIFY_DATA_ACCESS,
)
from django.conf import settings
from core_main_app.settings import DATA_SORTING_FIELDS
from core_main_app.utils.raw_query import django_raw_query
from core_main_app.utils.raw_query import mongo_raw_query

logger = logging.getLogger(__name__)


def has_perm_publish_data(user):
    """Does the user have the permission to publish a data.

    Args:
        user

    Returns
    """
    has_perm_publish(user, rights.PUBLISH_DATA)


def can_read_list_data_id(func, list_data_id, user):
    """Can read list of data.

    Args:
        func:
        list_data_id:
        user:

    Returns:

    """
    if user.is_superuser:
        return func(list_data_id, user)

    # check anonymous access
    check_anonymous_access(user)

    list_data = func(list_data_id, user)
    check_can_read_list(list_data, user)

    return list_data


def can_read_data_query(
    func,
    query,
    user,
    workspace_filter=None,
    user_filter=None,
    order_by_field=DATA_SORTING_FIELDS,
):
    """Can read a data, given a query.

    Args:
        func:
        query:
        user:
        workspace_filter:
        user_filter:
        order_by_field:

    Returns:

    """
    # check anonymous access
    check_anonymous_access(user)
    # update the query
    query = _update_can_read_query(query, user, workspace_filter, user_filter)
    # get list of data
    data_list = func(query, user, order_by_field=order_by_field)
    # if superuser, return list of data
    if user.is_superuser:
        return data_list
    # TODO: check if necessary because it is time consuming (checking that user has access to list of returned data)
    # check that user can access the list of data
    if VERIFY_DATA_ACCESS:
        check_can_read_list(data_list, user)
    return data_list


def can_read_aggregate_query(func, query, user):
    """Can read a data, given an aggregate query.

    Args:
        func:
        query:
        user:

    Returns:

    """
    if user.is_superuser:
        return func(query, user)

    # check anonymous access
    check_anonymous_access(user)

    # update the query
    query = _update_can_read_aggregate_query(query, user)

    return func(query, user)


def _update_can_read_query(
    query, user, workspace_filter=None, user_filter=None
):
    """Update query with access control parameters.

    Args:
        query:
        user:

    Returns:

    """

    accessible_workspaces = _get_read_accessible_workspaces_by_user(user)
    # update query with workspace criteria
    query = django_raw_query.add_access_criteria(
        query, accessible_workspaces, user, workspace_filter, user_filter
    )
    return query


def _update_can_read_aggregate_query(query, user):
    """Update query with access control parameters.

    Args:
        query:
        user:

    Returns:

    """

    accessible_workspaces = _get_read_accessible_workspaces_by_user(user)
    # update query with workspace criteria
    query = mongo_raw_query.add_aggregate_access_criteria(
        query, accessible_workspaces, user
    )
    return query


def _get_read_accessible_workspaces_by_user(user):
    """Get read accessible workspaces by user.

    Args:
        user:

    Returns:

    """
    if not settings.CAN_ANONYMOUS_ACCESS_PUBLIC_DOCUMENT and user.is_anonymous:
        accessible_workspaces = []
    else:
        # workspace case
        # list accessible workspaces
        accessible_workspaces = [
            workspace.id
            for workspace in workspace_api.get_all_workspaces_with_read_access_by_user(
                user
            )
        ]

    return accessible_workspaces


def can_write_data_workspace(func, data, workspace, user):
    """Can user write data in workspace.

    Args:
        func:
        data:
        workspace:
        user:

    Returns:

    """
    return can_write_in_workspace(
        func, data, workspace, user, rights.PUBLISH_DATA
    )


def can_read_blob(func, data, user):
    """Can read from object id.

    Args:
        func:
        data:
        user:

    Returns:

    """
    if user.is_superuser:
        return func(data, user)

    if data._blob is not None:
        check_can_read_document(data._blob, user)

    return func(data, user)



""" Unit test views
"""
from unittest.mock import patch, MagicMock

from django.contrib.sessions.middleware import SessionMiddleware
from django.test import RequestFactory, SimpleTestCase, override_settings

from core_main_app.access_control.exceptions import AccessControlError
from core_main_app.commons.exceptions import DoesNotExist
from core_main_app.components.data.models import Data
from core_main_app.components.template.models import Template
from core_main_app.components.user_preferences.models import UserPreferences
from core_main_app.utils import xml
from core_main_app.utils.tests_tools.MockUser import create_mock_user
from core_main_app.views.common.views import DataContentEditor, XmlEditor
from core_main_app.views.user.views import (
    set_timezone,
    custom_login,
    custom_logout,
)


class TestXmlEditorGenerateView(SimpleTestCase):
    """Test Xml Editor Generate View"""

    def setUp(self):
        """setUp

        Returns:

        """
        self.factory = RequestFactory()
        self.user1 = create_mock_user(user_id="1")

    @override_settings(INSTALLED_APPS=["core_curate_app", "core_parser_app"])
    @patch(
        "core_curate_app.components.curate_data_structure.models.CurateDataStructure"
    )
    @patch("core_main_app.utils.xml.format_content_xml")
    @patch("core_curate_app.views.user.views.render_xml")
    @patch("core_parser_app.components.data_structure_element.api.get_by_id")
    @patch("core_parser_app.tools.parser.parser.XSDParser.generate_form")
    @patch("core_curate_app.components.curate_data_structure.api.delete")
    @patch("core_curate_app.components.curate_data_structure.api.upsert")
    @patch("core_main_app.components.template.api.get_by_id")
    def test_generate_returns_response_with_xml(
        self,
        mock_template_get_by_id,
        mock_curate_ds_upsert,
        mock_curate_ds_delete,
        mock_generate_form,
        mock_dse_get_by_id,
        mock_render_xml,
        mock_format_xml,
        mock_curate_ds,
    ):
        """test_generate_with_content_returns_error

        Returns:

        """
        # Arrange
        mock_template_get_by_id.return_value = MagicMock(content="")
        mock_curate_ds_upsert.return_value = None
        mock_curate_ds_delete.return_value = None
        mock_generate_form.return_value = 1
        mock_dse_get_by_id.return_value = MagicMock()
        mock_render_xml.return_value = "<root></root>"
        mock_format_xml.return_value = "<root/>"
        mock_curate_ds.return_value = MagicMock()

        request = self.factory.get("core_main_app_xml_text_editor_view")
        request.user = self.user1
        data = {
            "content": "",
            "template_id": "1",
            "action": "generate",
        }
        request = self.factory.post("core_main_app_xml_text_editor_view", data)
        request.user = self.user1

        # Act
        response = DataContentEditor.as_view()(request)

        # Assert
        self.assertEqual(response.status_code, 200)
        self.assertTrue(mock_template_get_by_id.called)
        self.assertTrue(mock_curate_ds_upsert.called)
        self.assertTrue(mock_generate_form.called)
        self.assertTrue(mock_dse_get_by_id.called)
        self.assertTrue(mock_render_xml.called)
        self.assertTrue(mock_format_xml.called)
        self.assertTrue(mock_curate_ds.called)

    @override_settings(INSTALLED_APPS=["core_curate_app"])
    def test_generate_with_content_returns_error(self):
        """test_generate_with_content_returns_error

        Returns:

        """
        request = self.factory.get("core_main_app_xml_text_editor_view")
        request.user = self.user1
        data = {
            "content": "<root></root>",
            "template_id": "1",
            "action": "generate",
        }
        request = self.factory.post("core_main_app_xml_text_editor_view", data)
        request.user = self.user1
        response = DataContentEditor.as_view()(request)
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.content,
            b"Please clear form before generating a new XML document.",
        )

    def test_generate_without_core_curate_app_returns_error(self):
        """test_generate_without_core_curate_app_returns_error

        Returns:

        """
        request = self.factory.get("core_main_app_xml_text_editor_view")
        request.user = self.user1
        data = {
            "content": "",
            "template_id": "1",
            "action": "generate",
        }
        request = self.factory.post("core_main_app_xml_text_editor_view", data)
        request.user = self.user1
        response = DataContentEditor.as_view()(request)
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.content,
            b"The Curate App needs to be installed to use this feature.",
        )


class TestSetTimezone(SimpleTestCase):
    """TestSetTimezone"""

    def setUp(self):
        """setUp

        Returns:

        """
        self.factory = RequestFactory()
        self.user1 = create_mock_user(user_id="1")

    @patch("core_main_app.components.user_preferences.api.get_by_user")
    def test_get_when_user_has_preferences_returns_form(
        self, mock_get_by_user
    ):
        """test_get_when_user_has_preferences_returns_form

        Returns:

        """
        # Arrange
        mock_get_by_user.return_value = UserPreferences(
            user_id="1", timezone="GMT"
        )
        request = self.factory.get("core_main_set_timezone")
        request.user = self.user1

        # Act
        response = set_timezone(request)

        # Assert
        self.assertTrue("form" in response.content.decode())
        self.assertTrue(
            '<option value="GMT" selected>GMT</option>'
            in response.content.decode()
        )

    @patch("core_main_app.components.user_preferences.api.get_by_user")
    def test_get_returns_form(self, mock_get_by_user):
        """test_get_returns_form

        Returns:

        """
        # Arrange
        mock_get_by_user.return_value = UserPreferences(
            user_id="1", timezone="Africa/Casablanca"
        )
        request = self.factory.get("core_main_set_timezone")
        request.user = self.user1

        # Act
        response = set_timezone(request)

        # Assert
        self.assertTrue("form" in response.content.decode())
        self.assertTrue(
            '<option value="Africa/Casablanca" selected>Africa/Casablanca</option>'
            in response.content.decode()
        )

    @patch("core_main_app.components.user_preferences.api.get_by_user")
    @patch("core_main_app.components.user_preferences.api.upsert")
    def test_post_update_redirects(self, mock_upsert, mock_get_by_user):
        """test_post_update_redirects

        Args:
            mock_get_by_user:
            mock_upsert:

        Returns:

        """
        mock_get_by_user.return_value = UserPreferences(
            user_id="1", timezone="GMT"
        )
        mock_upsert.return_value = UserPreferences(user_id="1", timezone="UTC")

        # Arrange
        request = self.factory.post("core_main_set_timezone")
        request.POST = {"timezone": "UTC"}
        request.user = self.user1
        # Add middlewares
        mock_get_response = MagicMock()
        middleware = SessionMiddleware(mock_get_response)
        middleware.process_request(request)

        # Act
        response = set_timezone(request)

        # Assert
        self.assertEqual(response.status_code, 302)

    @patch("core_main_app.components.user_preferences.api.get_by_user")
    @patch("core_main_app.components.user_preferences.api.upsert")
    def test_post_new_redirects(self, mock_upsert, mock_get_by_user):
        """test_post_new_redirects

        Args:
            mock_get_by_user:
            mock_upsert:

        Returns:

        """
        mock_get_by_user.side_effect = DoesNotExist("error")
        mock_upsert.return_value = UserPreferences(user_id="1", timezone="UTC")

        # Arrange
        request = self.factory.post("core_main_set_timezone")
        request.POST = {"timezone": "UTC"}
        request.user = self.user1
        # Add middlewares
        mock_get_response = MagicMock()
        middleware = SessionMiddleware(mock_get_response)
        middleware.process_request(request)

        # Act
        response = set_timezone(request)

        # Assert
        self.assertEqual(response.status_code, 302)

    @patch("core_main_app.components.user_preferences.api.get_by_user")
    def test_get_acl_error_returns_error(self, mock_get_by_user):
        # Arrange
        mock_get_by_user.side_effect = AccessControlError("error")
        request = self.factory.post("core_main_set_timezone")
        request.user = create_mock_user("1")

        # Act
        response = set_timezone(request)

        # Assert
        self.assertTrue("Access Forbidden" in response.content.decode())


class TestDefaultCustomLogin(SimpleTestCase):
    """TestDefaultCustomLogin"""

    def setUp(self):
        """setUp

        Returns:

        """
        self.factory = RequestFactory()
        self.anonymous_user = create_mock_user(user_id=None, is_anonymous=True)
        self.connected_user = create_mock_user(user_id="1")
        self.inactive_user = create_mock_user(user_id=None, is_active=False)

    @override_settings(INSTALLED_APPS=["core_main_app"])
    @patch("core_main_app.components.web_page_login.api.get")
    def test_get_returns_login_form_and_forget_password_button(
        self, mock_web_page_login_get
    ):
        """test_post_inactive_user_shows_form_with_error

        Returns:

        """
        # Arrange
        request = self.factory.get("core_main_app_login")
        request.user = self.anonymous_user
        mock_web_page_login_get.return_value = None

        # Act
        response = custom_login(request)

        # Assert
        self.assertTrue("form" in response.content.decode())
        self.assertTrue("Forgot password" in response.content.decode())

    @override_settings(INSTALLED_APPS=["core_main_app"])
    @patch("core_main_app.components.web_page_login.api.get")
    @patch("django.contrib.auth.authenticate")
    @patch("django.contrib.auth.login")
    def test_post_inactive_user_shows_form_with_error(
        self,
        mock_login,
        mock_authenticate,
        mock_web_page_login_get,
    ):
        """test_post_inactive_user_shows_form_with_error

        Returns:

        """
        # Arrange
        request = self.factory.post("core_main_app_login")
        request.method = "POST"
        request.POST = {
            "username": "user",
            "password": "pass",
            "next_page": None,
        }
        request.user = self.inactive_user
        mock_web_page_login_get.return_value = None
        mock_authenticate.return_value = self.inactive_user
        mock_login.return_value = None

        # Act
        response = custom_login(request)

        # Assert
        self.assertTrue("form" in response.content.decode())
        self.assertTrue(
            "Your username is not activated yet." in response.content.decode()
        )

    @override_settings(INSTALLED_APPS=["core_main_app"])
    @patch("core_main_app.components.web_page_login.api.get")
    @patch("django.contrib.auth.authenticate")
    @patch("django.contrib.auth.login")
    def test_post_active_user_redirects_to_page(
        self,
        mock_login,
        mock_authenticate,
        mock_web_page_login_get,
    ):
        """test_post_inactive_user_shows_form_with_error

        Returns:

        """
        # Arrange
        request = self.factory.post("core_main_app_login")
        request.method = "POST"
        request.POST = {
            "username": "user",
            "password": "pass",
            "next_page": None,
        }
        request.user = self.inactive_user
        mock_web_page_login_get.return_value = None
        mock_authenticate.return_value = self.connected_user
        mock_login.return_value = None

        # Act
        response = custom_login(request)

        # Assert
        self.assertTrue(response.status_code, 302)

    @override_settings(INSTALLED_APPS=["core_main_app"])
    @patch("core_main_app.components.web_page_login.api.get")
    @patch("django.contrib.auth.authenticate")
    @patch("django.contrib.auth.login")
    def test_post_with_error_renders_login_page_with_errors(
        self,
        mock_login,
        mock_authenticate,
        mock_web_page_login_get,
    ):
        """test_post_inactive_user_shows_form_with_error

        Returns:

        """
        # Arrange
        request = self.factory.post("core_main_app_login")
        request.method = "POST"
        request.POST = {
            "username": "user",
            "password": "pass",
            "next_page": None,
        }
        request.user = self.inactive_user
        mock_web_page_login_get.return_value = None
        mock_authenticate.return_value = self.connected_user
        mock_login.side_effect = Exception()

        # Act
        response = custom_login(request)

        # Assert
        self.assertTrue("form" in response.content.decode())
        self.assertTrue(
            "Invalid username and/or password." in response.content.decode()
        )


class TestCustomLogout(SimpleTestCase):
    """TestCustomLogout"""

    def setUp(self):
        """setUp

        Returns:

        """
        self.factory = RequestFactory()
        self.connected_user = create_mock_user(user_id="1")

    @override_settings(INSTALLED_APPS=["core_main_app"])
    @patch("django.contrib.auth.logout")
    def test_logout_redirects(self, mock_logout):
        """test_logout_redirects

        Returns:

        """
        # Arrange
        request = self.factory.get("core_main_app_logout")
        request.user = self.connected_user
        mock_logout.return_value = None

        # Act
        response = custom_logout(request)

        # Assert
        self.assertTrue(response.status_code, 302)


class TestXmlEditor(SimpleTestCase):
    """Test Xml Editor"""

    @patch.multiple(XmlEditor, __abstractmethods__=set())
    @patch.object(xml, "format_content_xml")
    def test_get_context_calls_format_content_xml_when_content_is_not_empty(
        self, mock_format_content_xml
    ):
        """test_get_context_calls_format_content_xml_when_content_is_not_empty

        Returns:

        """
        # Arrange
        data = Data(template=Template())

        # Act
        XmlEditor().get_context(data, "title1", "<root>test</root>")

        # Assert
        mock_format_content_xml.assert_called()

    @patch.multiple(XmlEditor, __abstractmethods__=set())
    @patch.object(xml, "format_content_xml")
    def test_get_context_with_empty_content(self, mock_format_content_xml):
        """test_get_context_with_empty_content

        Returns:

        """
        # Arrange
        data = Data(template=Template())

        # Act
        XmlEditor().get_context(data, "title1", "")

        # Assert
        mock_format_content_xml.assert_not_called()



""" Custom admin site for the Database Lock model
"""
from django.contrib import admin


class CustomDatabaseLockAdmin(admin.ModelAdmin):
    """CustomDatabaseLockAdmin"""

    readonly_fields = ["object"]

    def has_add_permission(self, request, obj=None):
        """Prevent from manually adding Locks"""
        return False



"""Serializers for Xsl transformation
"""
from rest_framework.serializers import CharField, ModelSerializer

from core_main_app.commons.serializers import BasicSerializer
from core_main_app.components.xsl_transformation import api as xsl_api
from core_main_app.components.xsl_transformation.models import (
    XslTransformation,
)


class XslTransformationSerializer(ModelSerializer):
    """
    XslTransformation serializer
    """

    content = CharField(required=True)

    class Meta:
        """Meta"""

        model = XslTransformation
        fields = ["id", "name", "filename", "checksum", "content"]
        read_only_fields = ["id", "checksum"]

    def create(self, validated_data):
        """create

        Args:
            validated_data

        Returns:
        """
        return xsl_api.upsert(XslTransformation(**validated_data))

    def update(self, instance, validated_data):
        """update

        Args:
            instance:
            validated_data

        Returns:
        """
        instance.name = validated_data.get("name", instance.name)
        instance.content = validated_data.get("content", instance.content)
        instance.filename = validated_data.get("filename", instance.filename)
        return xsl_api.upsert(instance)


class TransformSerializer(BasicSerializer):
    """Transform Serializer"""

    xml_content = CharField(required=True)
    xslt_name = CharField(required=True)



""" Test forms.
"""
from unittest.case import TestCase
from unittest.mock import patch

from core_main_app.utils.tests_tools.MockUser import create_mock_user

from core_main_app.components.workspace.models import Workspace

from core_main_app.views.user.forms import ChangeWorkspaceForm

from core_main_app.views.admin.forms import TemplateXsltRenderingForm
from django.test import override_settings


class TestTemplateXsltRenderingForm(TestCase):
    """Test Template Xslt Rendering Form"""

    @override_settings(BOOTSTRAP_VERSION="4.6.2")
    def test_template_xslt_rendering_form_bootstrap_v4(self):
        """test_template_xslt_rendering_form_bootstrap_v4

        Returns:

        """
        data = {}
        form = TemplateXsltRenderingForm(data)
        self.assertEquals(
            form.fields["default_detail_xslt"].widget.attrs["class"],
            "form-control",
        )
        self.assertEquals(
            form.fields["list_xslt"].widget.attrs["class"], "form-control"
        )

    @override_settings(BOOTSTRAP_VERSION="5.1.3")
    def test_template_xslt_rendering_form_bootstrap_v5(self):
        """test_template_xslt_rendering_form_bootstrap_v5

        Returns:

        """
        data = {}
        form = TemplateXsltRenderingForm(data)
        self.assertEquals(
            form.fields["default_detail_xslt"].widget.attrs["class"],
            "form-select",
        )
        self.assertEquals(
            form.fields["list_xslt"].widget.attrs["class"], "form-select"
        )


class TestChangeWorkspaceForm(TestCase):
    """Test Change Workspace Form"""

    @override_settings(BOOTSTRAP_VERSION="4.6.2")
    @patch("core_main_app.components.user.api.get_user_by_id")
    @patch(
        "core_main_app.components.workspace.api.get_all_workspaces_with_write_access_by_user"
    )
    def test_change_workspace_form_bootstrap_v4(
        self, mock_get_all_workspaces_with_read_access_by_user, mock_user
    ):
        """test_change_workspace_form_bootstrap_v4

        Returns:

        """
        mock_get_all_workspaces_with_read_access_by_user.return_value = [
            Workspace(id=1)
        ]
        user = create_mock_user(user_id="1")
        user.username = "test"
        mock_user.return_value = user
        data = {"workspaces": list()}
        form = ChangeWorkspaceForm(data)
        self.assertEquals(
            form.fields["workspaces"].widget.attrs["class"], "form-control"
        )

    @override_settings(BOOTSTRAP_VERSION="5.1.3")
    @patch("core_main_app.components.user.api.get_user_by_id")
    @patch(
        "core_main_app.components.workspace.api.get_all_workspaces_with_write_access_by_user"
    )
    def test_change_workspace_form_bootstrap_v5(
        self, mock_get_all_workspaces_with_read_access_by_user, mock_user
    ):
        """test_change_workspace_form_bootstrap_v5

        Returns:

        """
        mock_get_all_workspaces_with_read_access_by_user.return_value = [
            Workspace(id=1)
        ]
        user = create_mock_user(user_id="1")
        user.username = "test"
        mock_user.return_value = user
        data = {"workspaces": list()}
        form = ChangeWorkspaceForm(data)
        self.assertEquals(
            form.fields["workspaces"].widget.attrs["class"], "form-select"
        )



""" Common views
"""
import json
from abc import ABCMeta, abstractmethod
from datetime import datetime

from django.conf import settings as conf_settings
from django.contrib import messages
from django.contrib.auth.decorators import login_required
from django.http import (
    HttpResponseBadRequest,
    HttpResponseForbidden,
    HttpResponse,
)
from django.http.response import HttpResponseRedirect
from django.urls import reverse
from django.utils.decorators import method_decorator
from django.utils.html import escape as html_escape
from django.views.generic import View

from core_main_app.access_control import api as acl_api
from core_main_app.access_control.exceptions import AccessControlError
from core_main_app.commons import exceptions
from core_main_app.commons.exceptions import (
    DoesNotExist,
)
from core_main_app.components.data import api as data_api
from core_main_app.components.group import api as group_api
from core_main_app.components.lock import api as lock_api
from core_main_app.components.template import api as template_api
from core_main_app.components.template.access_control import check_can_write
from core_main_app.components.template_xsl_rendering import (
    api as template_xsl_rendering_api,
)
from core_main_app.components.version_manager import api as version_manager_api
from core_main_app.components.workspace import api as workspace_api
from core_main_app.components.xsl_transformation import (
    api as xslt_transformation_api,
)
from core_main_app.settings import MAX_DOCUMENT_EDITING_SIZE
from core_main_app.utils import file as main_file_utils
from core_main_app.utils import group as group_utils
from core_main_app.utils import xml as main_xml_utils
from core_main_app.utils.labels import get_data_label
from core_main_app.utils.rendering import admin_render, render
from core_main_app.utils.view_builders import data as data_view_builder
from core_main_app.views.admin.forms import TemplateXsltRenderingForm
from xml_utils.xsd_tree.xsd_tree import XSDTree
from core_main_app.components.blob import api as blob_api


class CommonView(View, metaclass=ABCMeta):
    """
    Abstract common view for admin and user.
    """

    administration = False

    def common_render(
        self, request, template_name, modals=None, assets=None, context=None
    ):
        """common_render

        Args:
            request:
            template_name:
            modals:
            assets:
            context:

        Returns:
        """
        return (
            admin_render(request, template_name, modals, assets, context)
            if self.administration
            else render(request, template_name, modals, assets, context)
        )

    def is_administration(self):
        """is_administration

        Returns:
        """
        return self.administration


@method_decorator(login_required, name="dispatch")
class EditWorkspaceRights(CommonView):
    """
    Edit workspace rights
    """

    template = "core_main_app/user/workspaces/edit_rights.html"

    def get(self, request, *args, **kwargs):
        """get

        Args:
            request:

        Returns:
        """

        try:
            workspace_id = kwargs["workspace_id"]
            workspace = workspace_api.get_by_id(workspace_id)
        except DoesNotExist:
            return HttpResponseBadRequest("The workspace does not exist.")
        except Exception:
            return HttpResponseBadRequest("Something wrong happened.")

        if workspace.owner != str(request.user.id) and not self.administration:
            return HttpResponseForbidden(
                "Only the workspace owner can edit the rights."
            )

        try:
            # Users
            users_read_workspace = (
                workspace_api.get_list_user_can_read_workspace(
                    workspace, request.user
                )
            )
            users_write_workspace = (
                workspace_api.get_list_user_can_write_workspace(
                    workspace, request.user
                )
            )

            users_access_workspace = list(
                set(users_read_workspace + users_write_workspace)
            )
            detailed_users = []
            for user in users_access_workspace:
                if str(user.id) != workspace.owner:
                    detailed_users.append(
                        {
                            "object_id": user.id,
                            "object_name": user.username,
                            "can_read": user in users_read_workspace,
                            "can_write": user in users_write_workspace,
                        }
                    )
        except Exception:
            detailed_users = []

        try:
            # Groups
            groups_read_workspace = (
                workspace_api.get_list_group_can_read_workspace(
                    workspace, request.user
                )
            )
            groups_write_workspace = (
                workspace_api.get_list_group_can_write_workspace(
                    workspace, request.user
                )
            )

            groups_access_workspace = list(
                set(groups_read_workspace + groups_write_workspace)
            )
            group_utils.remove_list_object_from_list(
                groups_access_workspace,
                [
                    group_api.get_anonymous_group(),
                    group_api.get_default_group(),
                ],
            )
            detailed_groups = []
            for group in groups_access_workspace:
                detailed_groups.append(
                    {
                        "object_id": group.id,
                        "object_name": group.name,
                        "can_read": group in groups_read_workspace,
                        "can_write": group in groups_write_workspace,
                    }
                )
        except Exception:
            detailed_groups = []

        context = {
            "workspace": workspace,
            "user_data": detailed_users,
            "group_data": detailed_groups,
            "template": "core_main_app/user/workspaces/list/edit_rights_table.html",
            "action_read": "action_read",
            "action_write": "action_write",
        }

        if workspace_api.is_workspace_public(workspace):
            context.update({"is_public": True})
        if workspace_api.is_workspace_global(workspace):
            context.update({"is_global": True})

        assets = {
            "css": [
                "core_main_app/common/css/switch.css",
                "core_main_app/common/css/select.css",
            ],
            "js": [
                {
                    "path": "core_main_app/common/js/backtoprevious.js",
                    "is_raw": True,
                },
                {
                    "path": "core_main_app/user/js/workspaces/add_user.js",
                    "is_raw": False,
                },
                {
                    "path": "core_main_app/user/js/workspaces/list/modals/switch_right.js",
                    "is_raw": False,
                },
                {
                    "path": "core_main_app/user/js/workspaces/list/modals/remove_rights.js",
                    "is_raw": False,
                },
                {
                    "path": "core_main_app/user/js/workspaces/add_group.js",
                    "is_raw": False,
                },
                {
                    "path": "core_main_app/user/js/workspaces/init.js",
                    "is_raw": False,
                },
            ],
        }

        modals = [
            "core_main_app/user/workspaces/list/modals/add_user.html",
            "core_main_app/user/workspaces/list/modals/switch_right.html",
            "core_main_app/user/workspaces/list/modals/remove_rights.html",
            "core_main_app/user/workspaces/list/modals/add_group.html",
        ]

        # Set page title
        context.update({"page_title": "Edit Workspace"})

        return self.common_render(
            request,
            self.template,
            context=context,
            assets=assets,
            modals=modals,
        )


class ViewData(CommonView):
    """
    View detail data.
    """

    template = "core_main_app/user/data/detail.html"

    def get(self, request, *args, **kwargs):
        """get

        Args:
            request:

        Returns:
        """
        data_id = request.GET["id"]

        try:
            data_object = data_api.get_by_id(data_id, request.user)
            page_context = data_view_builder.build_page(
                data_object, display_download_options=True
            )

            return data_view_builder.render_page(
                request, self.common_render, self.template, page_context
            )
        except exceptions.DoesNotExist:
            error_message = "Data not found"
            status_code = 404
        except exceptions.ModelError:
            error_message = "Model error"
            status_code = 400
        except AccessControlError:
            error_message = "Access Forbidden"
            status_code = 403
        except Exception as e:
            error_message = str(e)
            status_code = 400

        return self.common_render(
            request,
            "core_main_app/common/commons/error.html",
            assets={
                "js": [
                    {
                        "path": "core_main_app/user/js/data/detail.js",
                        "is_raw": False,
                    }
                ]
            },
            context={
                "error": "Unable to access the requested "
                + get_data_label()
                + f": {error_message}.",
                "status_code": status_code,
                "page_title": "Error",
            },
        )


def read_xsd_file(xsd_file):
    """Return the content of the file uploaded using Django FileField.

    Returns:

    """
    # put the cursor at the beginning of the file
    xsd_file.seek(0)
    # read the content of the file
    return xsd_file.read().decode("utf-8")


@method_decorator(login_required, name="dispatch")
class TemplateXSLRenderingView(View):
    """Template XSL rendering view."""

    rendering = render
    save_redirect = "core_main_app_manage_template_versions"
    back_to_url = "core_main_app_manage_template_versions"
    form_class = TemplateXsltRenderingForm
    template_name = "core_main_app/common/templates_xslt/main.html"
    context = {}
    assets = {}

    def get(self, request, *args, **kwargs):
        """GET request. Create/Show the form for the configuration.

        Args:
            request:
            *args:
            **kwargs:

        Returns:

        """
        template_id = kwargs.pop("template_id")
        # Get the template
        template = template_api.get_by_id(template_id, request=request)
        # Get template information (version)
        version_manager = template.version_manager
        version_number = version_manager_api.get_version_number(
            version_manager, template_id, request=request
        )
        try:
            # Get the existing configuration to build the form
            template_xsl_rendering = (
                template_xsl_rendering_api.get_by_template_id(template_id)
            )
            data = {
                "id": str(template_xsl_rendering.id),
                "template": str(template.id),
                "list_xslt": template_xsl_rendering.list_xslt.id
                if template_xsl_rendering.list_xslt
                else None,
                "default_detail_xslt": template_xsl_rendering.default_detail_xslt.id
                if template_xsl_rendering.default_detail_xslt
                else None,
                "list_detail_xslt": [
                    xslt.id
                    for xslt in template_xsl_rendering.list_detail_xslt.all()
                ]
                if template_xsl_rendering.list_detail_xslt.count()
                else None,
            }
        except (Exception, exceptions.DoesNotExist):
            # If no configuration, new form with pre-selected fields.
            data = {
                "template": str(template.id),
                "list_xslt": None,
                "default_detail_xslt": None,
                "list_detail_xslt": None,
            }

        self.assets = {
            "css": ["core_main_app/admin/css/templates_xslt/form.css"],
            "js": [
                {
                    "path": "core_main_app/admin/js/templates_xslt/detail_xslt.js",
                    "is_raw": False,
                }
            ],
        }

        self.context = {
            "template_title": template.version_manager.title,
            "template_version": version_number,
            "form_template_xsl_rendering": self.form_class(data),
            "url_back_to": reverse(
                self.back_to_url,
                kwargs={"version_manager_id": template.version_manager.id},
            ),
        }

        return self.rendering(
            request,
            self.template_name,
            context=self.context,
            assets=self.assets,
        )

    def post(self, request, *args, **kwargs):
        """POST request. Try to save the configuration.

        Args:
            request:
            *args:
            **kwargs:

        Returns:

        """
        form = self.form_class(request.POST, request.FILES)
        self.context.update({"form_template_xsl_rendering": form})

        if form.is_valid():
            return self._save_template_xslt(request)
        else:
            # Display error from the form
            return self.rendering(
                request, self.template_name, context=self.context
            )

    def _save_template_xslt(self, request):
        """Save a template xslt rendering.

        Args:
            request: Request.

        """
        try:
            # Get the list xslt instance
            try:
                list_xslt = xslt_transformation_api.get_by_id(
                    request.POST.get("list_xslt")
                )
            except (Exception, exceptions.DoesNotExist):
                list_xslt = None

            # Get the list detail xslt instance
            try:
                list_detail_xslt = xslt_transformation_api.get_by_id_list(
                    request.POST.getlist("list_detail_xslt")
                )
            except (Exception, exceptions.DoesNotExist):
                list_detail_xslt = None

            # Get the default detail xslt instance
            try:
                default_detail_xslt = xslt_transformation_api.get_by_id(
                    request.POST.get("default_detail_xslt")
                )
            except (Exception, exceptions.DoesNotExist):
                default_detail_xslt = None

            # Get template by id
            template = template_api.get_by_id(
                request.POST.get("template"), request=request
            )

            template_xsl_rendering_api.add_or_delete(
                template_xsl_rendering_id=request.POST.get("id"),
                template=template,
                list_xslt=list_xslt,
                default_detail_xslt=default_detail_xslt,
                list_detail_xslt=list_detail_xslt,
            )

            return HttpResponseRedirect(
                reverse(self.save_redirect, args=[template.version_manager.id])
            )
        except Exception as e:
            self.context.update({"errors": html_escape(str(e))})
            return self.rendering(
                request, self.template_name, context=self.context
            )


def defender_error_page(request):
    """Error page for defender package.

    Args:
        request:

    Returns:

    """
    return render(
        request,
        "core_main_app/common/defender/error.html",
        context={"page_title": "Error"},
    )


class AbstractEditorView(View, metaclass=ABCMeta):
    """Abstract Text Editor View"""

    template = "core_main_app/user/text_editor/text_editor.html"

    def _get_assets(self):
        """get assets

        Return:
        """
        assets = {
            "js": [
                {
                    "path": "core_main_app/user/js/text_editor/text_editor.js",
                    "is_raw": True,
                },
            ],
            "css": [
                "core_main_app/common/css/highlight.css",
                "core_main_app/user/css/text-editor.css",
            ],
        }
        return assets

    def _get_context(self, document_id, document_name, type_content, content):
        """get context

        Return:
        """
        context = {
            "page_title": type_content + " Text Editor",
            "name": document_name,
            "content": content,
            "type": type_content,
            "document_id": document_id,
        }
        return context

    def post(self, request):
        """post

        Parameters:
            {
                "content":"content",
                "action": format/validate/save
                "document_id": document_id
                "template_id": template_id
            }

        Args:
            request: HTTP request

        Returns:

            - code: 200
              content: action done
            - code: 400
              content: Bad request
            - code: 403
              content: Forbidden
            - code: 500
              content: Internal server error
        """

        try:
            # get action
            action = request.POST["action"]
            # apply action: format, validate or save
            return getattr(self, "%s" % action)()
        except Exception as e:
            return HttpResponseBadRequest(html_escape(str(e)))

    @abstractmethod
    def format(self, *args, **kwargs):
        """Returns formatted content

        Args:
            args:
            kwargs:

        Returns: content

        """
        raise NotImplementedError("format method is not implemented.")

    @abstractmethod
    def validate(self, *args, **kwargs):
        """Validate content

        Args:
            args:
            kwargs:

        Returns:

        """
        raise NotImplementedError("validate method is not implemented.")

    @abstractmethod
    def save(self, *args, **kwargs):
        """Save content

        Args:
            args:
            kwargs:

        Returns:

        """
        raise NotImplementedError("save method is not implemented.")


class XmlEditor(AbstractEditorView, metaclass=ABCMeta):
    """Xml Editor"""

    save_redirect = "core_dashboard_records"

    def format(self, *args, **kwargs):
        """Format xml content

        Args:
            args:
            kwargs:

        Returns:

        """
        content = self.request.POST["content"].strip()
        return HttpResponse(
            json.dumps(main_xml_utils.format_content_xml(content)),
            "application/javascript",
        )

    def validate(self, *args, **kwargs):
        """Validate xml content

        Args:
            args:
            kwargs:

        Returns:

        """
        content = self.request.POST["content"].strip()

        try:
            # build the xsd tree
            xml_tree = XSDTree.build_tree(content)
        except Exception as exception:
            raise exceptions.XMLError(str(exception))
        try:
            # get template
            template_id = self.request.POST["template_id"]
            template = template_api.get_by_id(template_id, self.request)
            # build the xsd tree
            xsd_tree = XSDTree.build_tree(template.content)
        except Exception as exception:
            raise exceptions.XSDError(str(exception))

        # validate content
        error = main_xml_utils.validate_xml_data(
            xsd_tree, xml_tree, request=self.request
        )
        if error is not None:
            raise exceptions.XMLError(error)
        return HttpResponse(
            json.dumps("Validated successfully"),
            "application/javascript",
        )

    def generate(self, *args, **kwargs):
        """Generate xml content

        Args:
            args:
            kwargs:

        Returns:

        """
        content = self.request.POST["content"].strip()
        template_id = self.request.POST["template_id"]

        if "core_curate_app" not in conf_settings.INSTALLED_APPS:
            raise exceptions.CoreError(
                "The Curate App needs to be installed to use this feature."
            )

        if content:
            raise exceptions.XMLError(
                "Please clear form before generating a new XML document."
            )

        from core_main_app.utils.parser import get_parser
        from core_parser_app.components.data_structure_element import (
            api as data_structure_element_api,
        )
        from core_curate_app.components.curate_data_structure.models import (
            CurateDataStructure,
        )
        from core_curate_app.components.curate_data_structure import (
            api as curate_data_structure_api,
        )
        from core_curate_app.views.user import views as curate_views

        # Get template
        template = template_api.get_by_id(template_id, self.request)
        # Create temp data structure
        curate_data_structure = CurateDataStructure(
            user=str(self.request.user.id),
            template=template,
            name="text_editor_tmp_" + str(datetime.now()),
        )
        # create new curate data structure
        curate_data_structure_api.upsert(
            curate_data_structure, self.request.user
        )
        # build parser
        parser = get_parser(request=self.request)
        # generate form
        root_element_id = parser.generate_form(
            xsd_doc_data=template.content,
            xml_doc_data=None,
            data_structure=curate_data_structure,
            request=self.request,
        )
        # get the root element
        root_element = data_structure_element_api.get_by_id(
            root_element_id, self.request
        )

        # generate xml string
        xml_data = curate_views.render_xml(self.request, root_element)

        # prettify content
        xml_data = main_xml_utils.format_content_xml(xml_data)

        # delete temp data structure
        curate_data_structure_api.delete(
            curate_data_structure, self.request.user
        )

        return HttpResponse(
            json.dumps(xml_data),
            "application/javascript",
        )

    def _get_assets(self):
        """get assets

        Return:
        """
        # get assets
        assets = super()._get_assets()

        # add css relatives to xml editor
        assets["css"].append("core_main_app/common/css/XMLTree.css")

        # add js relatives to xml editor
        assets["js"].append(
            {
                "path": "core_main_app/common/js/XMLTree.js",
                "is_raw": False,
            }
        )
        assets["js"].append(
            {
                "path": "core_main_app/user/js/text_editor/data_text_editor.raw.js",
                "is_raw": True,
            }
        )
        return assets

    def get_context(self, document, document_name, xml_content):
        """get context

        Args:
            document:
            document_name:
            xml_content:

        Returns:
        """
        # format before build context
        if xml_content != "":
            xml_content = main_xml_utils.format_content_xml(xml_content)

        # get assets
        context = super()._get_context(
            document.id, document_name, "XML", xml_content
        )

        # build xslt selector
        (
            display_xslt_selector,
            template_xsl_rendering,
            xsl_transformation_id,
        ) = data_view_builder.xslt_selector(document.template.id)

        # add context relatives to xml editor
        context.update(
            {
                "document_name": document.__class__.__name__,
                "template_id": document.template.id,
                "template_xsl_rendering": template_xsl_rendering,
                "xsl_transformation_id": xsl_transformation_id,
                "can_display_selector": display_xslt_selector,
            }
        )

        return context


class DataContentEditor(XmlEditor):
    """Data Content Editor View"""

    def get(self, request):
        """get

        Args:
            request

        Returns:
        """

        try:
            data = data_api.get_by_id(request.GET["id"], request.user)
            acl_api.check_can_write(data, request.user)
            if (
                main_file_utils.get_byte_size_from_string(data.xml_content)
                > MAX_DOCUMENT_EDITING_SIZE
            ):
                raise exceptions.DocumentEditingSizeError(
                    "The file is too large (MAX_DOCUMENT_EDITING_SIZE)."
                )
            lock_api.set_lock_object(data, request.user)
            context = self.get_context(data, data.title, data.xml_content)
            assets = self._get_assets()
            return render(
                request, self.template, assets=assets, context=context
            )
        except AccessControlError as acl_exception:
            error_message = "Access Forbidden: " + str(acl_exception)
            status_code = 403
        except exceptions.DoesNotExist:
            error_message = get_data_label() + " not found"
            status_code = 404
        except Exception as e:
            error_message = str(e)
            status_code = 400

        return render(
            request,
            "core_main_app/common/commons/error.html",
            assets={
                "js": [
                    {
                        "path": "core_main_app/user/js/data/detail.js",
                        "is_raw": False,
                    }
                ]
            },
            context={
                "error": error_message,
                "status_code": status_code,
            },
        )

    def save(self, *args, **kwargs):
        """Save xml content

        Args:
            args:
            kwargs:

        Returns:

        """
        try:
            content = self.request.POST["content"].strip()
            data_id = self.request.POST["document_id"]
            data = data_api.get_by_id(data_id, self.request.user)
            # update content
            data.xml_content = content
            # save data
            data_api.upsert(data, self.request)
            lock_api.remove_lock_on_object(data, self.request.user)
            messages.add_message(
                self.request,
                messages.SUCCESS,
                get_data_label().capitalize() + " saved with success.",
            )
            return HttpResponse(
                json.dumps({"url": reverse(self.save_redirect)}),
                "application/javascript",
            )
        except AccessControlError as ace:
            return HttpResponseForbidden(html_escape(str(ace)))
        except DoesNotExist as dne:
            return HttpResponseBadRequest(html_escape(str(dne)))
        except Exception as e:
            return HttpResponseBadRequest(html_escape(str(e)))


class XSDEditor(AbstractEditorView):
    """XSD Editor View"""

    def get(self, request):
        """get

        Args:
            request:

        Returns:
        """

        try:
            template = template_api.get_by_id(request.GET["id"], request)
            check_can_write(template, request=request)
            context = super()._get_context(
                template.id, template.filename, "XSD", template.content
            )
            assets = super()._get_assets()
            # add js relatives to template editor
            assets["js"].append(
                {
                    "path": "core_main_app/user/js/text_editor/template_text_editor.raw.js",
                    "is_raw": True,
                },
            )
            return render(
                request, self.template, assets=assets, context=context
            )

        except AccessControlError as acl_exception:
            error_message = str(acl_exception)
            status_code = 403
        except exceptions.DoesNotExist:
            error_message = "Template not found"
            status_code = 404
        except Exception as e:
            error_message = str(e)
            status_code = 400
        return render(
            request,
            "core_main_app/common/commons/error.html",
            assets={
                "js": [
                    {
                        "path": "core_main_app/user/js/data/detail.js",
                        "is_raw": False,
                    }
                ]
            },
            context={
                "error": error_message,
                "status_code": status_code,
            },
        )

    def format(self, *args, **kwargs):
        """Format xml content

        Args:
            args:
            kwargs:

        Returns:

        """
        content = self.request.POST["content"].strip()
        return HttpResponse(
            json.dumps(main_xml_utils.format_content_xml(content)),
            "application/javascript",
        )

    def validate(self, *args, **kwargs):
        """Validate xml content

        Args:
            args:
            kwargs:

        Returns:

        """
        content = self.request.POST["content"].strip()
        try:
            # build the xsd tree
            xsd_tree = XSDTree.build_tree(content)
        except Exception as exception:
            raise exceptions.XSDError(str(exception))

        # validate the schema
        error = main_xml_utils.validate_xml_schema(
            xsd_tree, request=self.request
        )
        if error is not None:
            raise exceptions.XMLError(error)

        return HttpResponse(
            json.dumps("Validated successfully"),
            "application/javascript",
        )

    def save(self, *args, **kwargs):
        """Save content

        Args:
            args:
            kwargs:

        Returns:

        """
        try:
            content = self.request.POST["content"].strip()
            template_id = self.request.POST["document_id"]
            template = template_api.get_by_id(template_id, self.request)
            # update content
            template.content = content
            # save template
            template_api.upsert(template, request=self.request)
            return HttpResponse(
                json.dumps("saved successfully"),
                "application/javascript",
            )
        except AccessControlError as ace:
            return HttpResponseForbidden(html_escape(str(ace)))
        except DoesNotExist as dne:
            return HttpResponseBadRequest(html_escape(str(dne)))
        except Exception as e:
            return HttpResponseBadRequest(html_escape(str(e)))


class ViewBlob(CommonView):
    """
    View blob.
    """

    template = "core_main_app/user/blob/detail.html"

    def get(self, request, *args, **kwargs):
        """get

        Args:
            request:

        Returns:
        """
        blob_id = request.GET["id"]

        try:
            # Get blob
            blob_object = blob_api.get_by_id(blob_id, request.user)
            try:
                acl_api.check_can_write(blob_object, request.user)
                can_write = True
            except AccessControlError:
                can_write = False
            # Init context
            context = {"blob": blob_object, "can_write": can_write}
            # Init assets
            assets = {
                "js": [
                    {
                        "path": "core_main_app/user/js/blob/detail.js",
                        "is_raw": False,
                    }
                ],
                "css": [],
            }

            context.update({"page_title": "View File"})

            return self.common_render(
                request,
                self.template,
                assets=assets,
                context=context,
            )
        except exceptions.DoesNotExist:
            error_message = "Blob not found"
            status_code = 404
        except AccessControlError:
            error_message = "Access Forbidden"
            status_code = 403
        except Exception as e:
            error_message = str(e)
            status_code = 400

        return self.common_render(
            request,
            "core_main_app/common/commons/error.html",
            context={
                "error": "Unable to access the requested file"
                + f": {error_message}.",
                "status_code": status_code,
                "page_title": "Error",
            },
        )


class ManageBlobMetadata(CommonView):
    """
    Manage blob metadata.
    """

    template = "core_main_app/user/blob/blob_metadata.html"

    def get(self, request, pk, *args, **kwargs):
        """get

        Args:
            request:
            pk:

        Returns:
        """

        try:
            # Get blob
            blob_object = blob_api.get_by_id(pk, request.user)
            try:
                acl_api.check_can_write(blob_object, request.user)
                can_write = True
            except AccessControlError:
                can_write = False
            # Init context
            context = {"blob": blob_object, "can_write": can_write}
            # Init modals
            modals = [
                "core_main_app/user/blob/list/modals/add_metadata.html",
            ]
            # Init assets
            assets = {
                "js": [
                    {
                        "path": "core_main_app/user/js/blob/detail.js",
                        "is_raw": False,
                    },
                    {
                        "path": "core_main_app/user/js/blob/blob_metadata.js",
                        "is_raw": False,
                    },
                ],
                "css": [
                    "core_main_app/common/css/select.css",
                ],
            }

            context.update({"page_title": "File Metadata"})

            return self.common_render(
                request,
                self.template,
                modals=modals,
                assets=assets,
                context=context,
            )
        except exceptions.DoesNotExist:
            error_message = "Blob not found"
            status_code = 404
        except AccessControlError:
            error_message = "Access Forbidden"
            status_code = 403
        except Exception as e:
            error_message = str(e)
            status_code = 400

        return self.common_render(
            request,
            "core_main_app/common/commons/error.html",
            context={
                "error": "Unable to access the requested file "
                + f": {error_message}.",
                "status_code": status_code,
                "page_title": "Error",
            },
        )



""" Mongoengine Data model
"""

import logging
from django.conf import settings

from core_main_app.commons.exceptions import CoreError
from core_main_app.components.data.models import Data
from core_main_app.components.data.tasks import (
    index_mongo_data,
    delete_mongo_data,
    update_mongo_data_user,
    update_mongo_data_workspace,
)
from core_main_app.components.template.models import Template
from core_main_app.components.workspace.models import Workspace
from core_main_app.settings import (
    SEARCHABLE_DATA_OCCURRENCES_LIMIT,
    XML_POST_PROCESSOR,
    XML_FORCE_LIST,
)
from core_main_app.utils import xml as xml_utils

logger = logging.getLogger(__name__)

try:
    if settings.MONGODB_INDEXING:
        from bson import ObjectId
        from mongoengine import Document, DoesNotExist
        from mongoengine import fields as mongo_fields

        class AbstractMongoData(Document):
            """Data object stored in MongoDB"""

            data_id = mongo_fields.IntField(primary_key=True)
            title = mongo_fields.StringField()
            dict_content = mongo_fields.DictField()
            creation_date = mongo_fields.DateTimeField()
            last_modification_date = mongo_fields.DateTimeField()
            last_change_date = mongo_fields.DateTimeField()
            mongo_id = mongo_fields.ObjectIdField()

            meta = {
                "abstract": True,
            }

            def get_dict_content(self):
                """Return dict_content

                Returns:

                """
                return self.dict_content

            @staticmethod
            def post_save_data(sender, instance, **kwargs):
                """post_save_data

                Args:
                    sender:
                    instance:
                    kwargs:

                Returns:

                """
                raise NotImplementedError("post_save_data not implemented")

            @staticmethod
            def post_delete_data(sender, instance, **kwargs):
                """post_delete_data

                Args:
                    sender:
                    instance:
                    kwargs:

                Returns:

                """
                raise NotImplementedError("post_delete_data not implemented")

        class MongoData(AbstractMongoData):
            """Data object stored in MongoDB"""

            _template_id = mongo_fields.IntField(db_field="template")
            user_id = mongo_fields.IntField()
            _workspace_id = mongo_fields.IntField(db_field="workspace")
            _xml_content = None

            meta = {
                "indexes": [
                    "title",
                    "_template_id",
                    "user_id",
                    "last_modification_date",
                ],
            }

            @property
            def template(self):
                """Return template object

                Returns:

                """
                return Template.get_by_id(self._template_id)

            @property
            def _blob(self):
                """Return blob object

                Returns:

                """
                try:
                    return Data.get_by_id(self.data_id)._blob
                except Exception:
                    logger.error("Unable to get MongoData._blob")
                    return None

            def blob(self, user):
                """Return blob object

                Args:
                    user:

                Returns:

                """
                return Data.get_by_id(self.data_id).blob(user)

            @property
            def template_id(self):
                """Return template id

                Returns:

                """
                return self._template_id

            @property
            def workspace(self):
                """Return workspace object

                Returns:

                """
                return (
                    Workspace.get_by_id(self._workspace_id)
                    if self._workspace_id
                    else None
                )

            @property
            def xml_content(self):
                """Get xml content - read from data.

                Returns:

                """
                if not self._xml_content:
                    self._xml_content = Data.get_by_id(
                        self.data_id
                    ).xml_content
                return self._xml_content

            @xml_content.setter
            def xml_content(self, value):
                """Set xml content - to be saved as a file.

                Args:
                    value:

                Returns:

                """
                self._xml_content = value

            @staticmethod
            def execute_query(query, order_by_field):
                """Execute a query.

                Args:
                    query:
                    order_by_field: Order by field.

                Returns:

                """
                return MongoData.objects.filter(query).order_by(
                    *order_by_field
                )

            @staticmethod
            def aggregate(pipeline):
                """Execute an aggregate on the Data collection.

                Args:
                    pipeline:

                Returns:

                """
                return MongoData.objects().aggregate(pipeline)

            @staticmethod
            def init_mongo_data(data):
                """Initialize mongo data from data

                Args:
                    data:

                Returns:

                """
                try:
                    # check if data already exists in mongo
                    mongo_data = MongoData.objects.get(pk=data.id)
                except DoesNotExist:
                    # create new mongo data otherwise
                    mongo_data = MongoData()
                    mongo_data.mongo_id = ObjectId()

                # Initialize mongo data fields
                mongo_data.data_id = data.id
                mongo_data.title = data.title
                # transform xml content into a dictionary
                mongo_data.dict_content = xml_utils.raw_xml_to_dict(
                    data.xml_content,
                    postprocessor=XML_POST_PROCESSOR,
                    force_list=XML_FORCE_LIST,
                    list_limit=SEARCHABLE_DATA_OCCURRENCES_LIMIT,
                )
                mongo_data._template_id = (
                    data.template.id if data.template else None
                )
                mongo_data.user_id = data.user_id if data.user_id else None
                mongo_data._workspace_id = (
                    data.workspace.id if data.workspace else None
                )
                mongo_data.creation_date = data.creation_date
                mongo_data.last_modification_date = data.last_modification_date
                mongo_data.last_change_date = data.last_change_date
                return mongo_data

            @staticmethod
            def update_user_id_from_queryset(data_queryset, user_id):
                """Update user id of all data in queryset

                Args:
                    data_queryset:
                    user_id:

                Returns:

                """
                data_ids = data_queryset.values_list("id", flat=True)
                if settings.MONGODB_ASYNC_SAVE:
                    update_mongo_data_user.apply_async(
                        (
                            list(data_ids),
                            user_id,
                        )
                    )
                else:
                    for data_id in data_ids:
                        mongo_data = MongoData.objects.get(pk=data_id)
                        mongo_data.user_id = user_id
                        mongo_data.save()

            @staticmethod
            def update_workspace_id_from_queryset(data_queryset, workspace_id):
                """Update workspace id of all data in queryset

                Args:
                    data_queryset:
                    workspace_id:

                Returns:

                """
                data_ids = data_queryset.values_list("id", flat=True)
                if settings.MONGODB_ASYNC_SAVE:
                    update_mongo_data_workspace.apply_async(
                        (
                            list(data_ids),
                            workspace_id,
                        )
                    )
                else:
                    for data_id in data_ids:
                        mongo_data = MongoData.objects.get(pk=data_id)
                        mongo_data._workspace_id = workspace_id
                        mongo_data.save()

            @staticmethod
            def post_save_data(sender, instance, **kwargs):
                """Method executed after saving of Data.
                Args:
                    sender: Class.
                    instance: Data document.
                    **kwargs: Args.

                """
                if settings.MONGODB_ASYNC_SAVE:
                    index_mongo_data.apply_async((str(instance.id),))
                else:
                    mongo_data = MongoData.init_mongo_data(instance)
                    mongo_data.save()

            @staticmethod
            def post_delete_data(sender, instance, **kwargs):
                """Method executed after deleting a Data.
                Args:
                    sender: Class.
                    instance: Data document.
                    **kwargs: Args.

                """
                if settings.MONGODB_ASYNC_SAVE:
                    delete_mongo_data.apply_async((str(instance.id),))
                else:
                    try:
                        mongo_data = MongoData.objects.get(data_id=instance.id)
                        mongo_data.delete()
                    except DoesNotExist:
                        logger.warning(
                            f"Trying to delete {str(instance.id)} but document was not found."
                        )
                    except Exception as e:
                        logger.error(f"An unexpected error occurred: {str(e)}")

            @staticmethod
            def pre_delete_workspace(sender, instance, **kwargs):
                """Method executed before deleting a Workspace.
                Args:
                    sender: Class.
                    instance: Workspace document.
                    **kwargs: Args.

                """
                queryset = Data.objects.filter(workspace=instance.id).all()
                MongoData.update_workspace_id_from_queryset(queryset, None)

except ImportError:
    raise CoreError(
        "Mongoengine needs to be installed when MongoDB indexing is enabled. "
        "Install required python packages (see requirements.mongo.txt) "
        "or disable MongoDB indexing (MONGODB_INDEXING=False). "
    )



""" Unit test for MongoData
"""
from unittest import TestCase
from unittest.mock import patch

from django.test import tag
from tests.components.data.tests_unit import (
    _create_data,
    _create_blob,
    _get_template,
)

from core_main_app.commons.exceptions import ModelError
from core_main_app.components.data.models import Data


class TestMongoDataBlob(TestCase):
    @tag("mongodb")
    @patch.object(Data, "get_by_id")
    def test_data__blob_returns_blob_if_exists(self, mock_get_by_id):
        """

        Returns:

        """
        from core_main_app.components.mongo.models import (
            MongoData,
        )

        blob = _create_blob(user_id="2")
        blob.save()
        template = _get_template()
        template.save()
        data = _create_data(
            data_id=1,
            template=template,
            user_id="2",
            title="new_title",
            content="<tag></tag>",
            blob=blob,
        )
        data.save()
        mock_get_by_id.return_value = data
        mongo_data = MongoData()
        self.assertEqual(mongo_data._blob, blob)

    @tag("mongodb")
    @patch.object(Data, "get_by_id")
    def test_data__blob_returns_none_if_error(self, mock_get_by_id):
        from core_main_app.components.mongo.models import (
            MongoData,
        )

        mock_get_by_id.side_effect = ModelError("error")
        mongo_data = MongoData()
        self.assertIsNone(mongo_data._blob)



""" Datetime utils
"""
from django.utils import timezone


def datetime_now():
    """Returns the current date

    Returns:

    """
    return timezone.now()


def datetime_timedelta(**kwargs):
    """Returns a timedelta

    Returns:

    """
    return timezone.timedelta(**kwargs)



""" Ajax API
"""
import json

from django.contrib import messages
from django.contrib.auth.decorators import login_required
from django.http import (
    HttpResponse,
    HttpResponseBadRequest,
    HttpResponseForbidden,
)
from django.template import loader
from django.utils.decorators import method_decorator
from django.utils.html import escape
from django.views.generic import View

from core_main_app.access_control.api import check_can_write
from core_main_app.access_control.exceptions import AccessControlError
from core_main_app.commons import exceptions
from core_main_app.commons.exceptions import DoesNotExist, ModelError
from core_main_app.components.blob import api as blob_api
from core_main_app.components.blob.models import Blob
from core_main_app.components.data import api as data_api
from core_main_app.components.group import api as group_api
from core_main_app.components.template import api as template_api
from core_main_app.components.user import api as user_api
from core_main_app.components.workspace import api as workspace_api
from core_main_app.templatetags.xsl_transform_tag import (
    render_xml_as_html_detail,
)
from core_main_app.utils import group as group_utils
from core_main_app.views.user.forms import (
    ChangeWorkspaceForm,
    UserRightForm,
    GroupRightForm,
    BlobMetadataForm,
    BlobFileForm,
)

GROUP = "group"
USER = "user"

ACTION_READ = "action_read"
ACTION_WRITE = "action_write"


@method_decorator(login_required, name="dispatch")
class LoadFormChangeWorkspace(View):
    """Load the form to list the workspaces."""

    show_global_workspace = True

    def post(self, request, *args, **kwargs):
        """post

        Args:
             request:

        Returns:

        """

        is_administration = request.POST.get("administration", False) == "True"

        try:
            form = ChangeWorkspaceForm(
                request.user,
                list(),
                is_administration,
                self.show_global_workspace,
            )
        except DoesNotExist as dne:
            return HttpResponseBadRequest(escape(str(dne)))
        except Exception:
            return HttpResponseBadRequest("Something wrong happened.")

        context = {"assign_workspace_form": form}

        return HttpResponse(
            json.dumps(
                {
                    "form": loader.render_to_string(
                        "core_main_app/user/workspaces/list/modals/assign_workspace_form.html",
                        context,
                    )
                }
            ),
            "application/javascript",
        )


@login_required
def load_add_user_form(request):
    """Load the form to list the users with no access to the workspace.

    Args:
        request:

    Returns:
    """
    workspace_id = request.POST.get("workspace_id", None)
    try:
        workspace = workspace_api.get_by_id(str(workspace_id))
    except exceptions.ModelError:
        return HttpResponseBadRequest("Invalid input.")
    except Exception:
        return HttpResponseBadRequest("An unexpected error occurred.")

    try:
        # We retrieve all users with no access
        users_with_no_access = list(
            workspace_api.get_list_user_with_no_access_workspace(
                workspace, request.user
            )
        )

        # We remove the owner of the workspace
        if len(users_with_no_access) > 0:
            users_with_no_access.remove(
                user_api.get_user_by_id(workspace.owner)
            )

        if len(users_with_no_access) == 0:
            return HttpResponseBadRequest(
                "There are no users that can be added."
            )

        form = UserRightForm(users_with_no_access)
    except AccessControlError as ace:
        return HttpResponseBadRequest(escape(str(ace)))
    except DoesNotExist as dne:
        return HttpResponseBadRequest(escape(str(dne)))
    except Exception:
        return HttpResponseBadRequest("Something wrong happened.")

    context = {"add_user_form": form}

    return HttpResponse(
        json.dumps(
            {
                "form": loader.render_to_string(
                    "core_main_app/user/workspaces/list/modals/add_user_form.html",
                    context,
                )
            }
        ),
        "application/javascript",
    )


@login_required
def add_user_right_to_workspace(request):
    """Add rights to user for the workspace.

    Args:
        request

    Returns
    """
    workspace_id = request.POST.get("workspace_id", None)
    users_ids = request.POST.getlist("users_id[]", [])
    is_read_checked = request.POST.get("read", None) == "true"
    is_write_checked = request.POST.get("write", None) == "true"

    if len(users_ids) == 0:
        return HttpResponseBadRequest("You need to select at least one user.")
    if not is_read_checked and not is_write_checked:
        return HttpResponseBadRequest(
            "You need to select at least one permission (read and/or write)."
        )

    try:
        workspace = workspace_api.get_by_id(str(workspace_id))
        for user in user_api.get_all_users_by_list_id(users_ids):
            if is_read_checked:
                workspace_api.add_user_read_access_to_workspace(
                    workspace, user, request.user
                )
            if is_write_checked:
                workspace_api.add_user_write_access_to_workspace(
                    workspace, user, request.user
                )
    except AccessControlError as ace:
        return HttpResponseBadRequest(escape(str(ace)))
    except DoesNotExist as dne:
        return HttpResponseBadRequest(escape(str(dne)))
    except Exception:
        return HttpResponseBadRequest("Something wrong happened.")

    return HttpResponse(json.dumps({}), content_type="application/javascript")


@login_required
def switch_right(request):
    """Switch user's right for the workspace.

    Args:
        request

    Returns
    """

    workspace_id = request.POST.get("workspace_id", None)
    object_id = request.POST.get("object_id", None)
    group_or_user = request.POST.get("group_or_user", None)
    action = request.POST.get("action", None)
    value = request.POST.get("value", None) == "true"

    try:
        workspace = workspace_api.get_by_id(str(workspace_id))

        if group_or_user == USER:
            _switch_user_right(
                object_id, action, value, workspace, request.user
            )
        if group_or_user == GROUP:
            _switch_group_right(
                object_id, action, value, workspace, request.user
            )

    except AccessControlError as ace:
        return HttpResponseBadRequest(escape(str(ace)))
    except DoesNotExist as dne:
        return HttpResponseBadRequest(escape(str(dne)))
    except Exception:
        return HttpResponseBadRequest("Something wrong happened.")

    return HttpResponse(json.dumps({}), content_type="application/javascript")


def _switch_user_right(user_id, action, value, workspace, request_user):
    """Change the user rights to the workspace.

    Args:
        user_id:
        action:
        value:
        workspace:
        request_user:

    Returns:
    """
    user = user_api.get_user_by_id(user_id)

    if action == ACTION_READ:
        if value:
            workspace_api.add_user_read_access_to_workspace(
                workspace, user, request_user
            )
        else:
            workspace_api.remove_user_read_access_to_workspace(
                workspace, user, request_user
            )
    elif action == ACTION_WRITE:
        if value:
            workspace_api.add_user_write_access_to_workspace(
                workspace, user, request_user
            )
        else:
            workspace_api.remove_user_write_access_to_workspace(
                workspace, user, request_user
            )


def _switch_group_right(group_id, action, value, workspace, request_user):
    """Change the group rights to the workspace.

    Args:
        group_id:
        action:
        value:
        workspace:
        request_user:

    Returns:
    """
    group = group_api.get_group_by_id(group_id)

    if action == ACTION_READ:
        if value:
            workspace_api.add_group_read_access_to_workspace(
                workspace, group, request_user
            )
        else:
            workspace_api.remove_group_read_access_to_workspace(
                workspace, group, request_user
            )
    elif action == ACTION_WRITE:
        if value:
            workspace_api.add_group_write_access_to_workspace(
                workspace, group, request_user
            )
        else:
            workspace_api.remove_group_write_access_to_workspace(
                workspace, group, request_user
            )


@login_required
def remove_user_or_group_rights(request):
    """Remove user's right for the workspace.

    Args:
        request

    Returns
    """

    workspace_id = request.POST.get("workspace_id", None)
    object_id = request.POST.get("object_id", None)
    group_or_user = request.POST.get("group_or_user", None)

    try:
        workspace = workspace_api.get_by_id(str(workspace_id))

        if group_or_user == USER:
            _remove_user_rights(object_id, workspace, request.user)
        if group_or_user == GROUP:
            _remove_group_rights(object_id, workspace, request.user)

    except AccessControlError as ace:
        return HttpResponseBadRequest(escape(str(ace)))
    except ModelError:
        return HttpResponseBadRequest("Invalid input.")
    except DoesNotExist as dne:
        return HttpResponseBadRequest(escape(str(dne)))
    except Exception:
        return HttpResponseBadRequest("Something wrong happened.")

    return HttpResponse(json.dumps({}), content_type="application/javascript")


def _remove_user_rights(object_id, workspace, request_user):
    """Remove all user rights on the workspace.

    Args:
        object_id:
        workspace:
        request_user:

    Returns:
    """
    user = user_api.get_user_by_id(object_id)
    workspace_api.remove_user_read_access_to_workspace(
        workspace, user, request_user
    )
    workspace_api.remove_user_write_access_to_workspace(
        workspace, user, request_user
    )


def _remove_group_rights(object_id, workspace, request_user):
    """Remove all group rights on the workspace.

    Args:
        object_id:
        workspace:
        request_user:

    Returns:
    """
    group = group_api.get_group_by_id(object_id)
    workspace_api.remove_group_read_access_to_workspace(
        workspace, group, request_user
    )
    workspace_api.remove_group_write_access_to_workspace(
        workspace, group, request_user
    )


@login_required
def load_add_group_form(request):
    """Load the form to list the groups with no access to the workspace.

    Args:
        request:

    Returns:
    """
    workspace_id = request.POST.get("workspace_id", None)
    try:
        workspace = workspace_api.get_by_id(str(workspace_id))
    except exceptions.ModelError:
        return HttpResponseBadRequest("Invalid input.")
    except Exception:
        return HttpResponseBadRequest("An unexpected error occurred.")

    try:
        # We retrieve all groups with no access
        groups_with_no_access = list(
            workspace_api.get_list_group_with_no_access_workspace(
                workspace, request.user
            )
        )

        if len(groups_with_no_access) > 0:
            group_utils.remove_list_object_from_list(
                groups_with_no_access,
                [
                    group_api.get_anonymous_group(),
                    group_api.get_default_group(),
                ],
            )
        if len(groups_with_no_access) == 0:
            return HttpResponseBadRequest(
                "There are no groups that can be added."
            )

        form = GroupRightForm(groups_with_no_access)
    except AccessControlError as ace:
        return HttpResponseBadRequest(escape(str(ace)))
    except DoesNotExist as dne:
        return HttpResponseBadRequest(escape(str(dne)))
    except Exception:
        return HttpResponseBadRequest("Something wrong happened.")

    context = {"add_group_form": form}

    return HttpResponse(
        json.dumps(
            {
                "form": loader.render_to_string(
                    "core_main_app/user/workspaces/list/modals/add_group_form.html",
                    context,
                )
            }
        ),
        "application/javascript",
    )


@login_required
def add_group_right_to_workspace(request):
    """Add rights to group for the workspace.

    Args:
        request

    Returns
    """
    workspace_id = request.POST.get("workspace_id", None)
    groups_ids = request.POST.getlist("groups_id[]", [])
    is_read_checked = request.POST.get("read", None) == "true"
    is_write_checked = request.POST.get("write", None) == "true"

    if len(groups_ids) == 0:
        return HttpResponseBadRequest("You need to select at least one group.")
    if not is_read_checked and not is_write_checked:
        return HttpResponseBadRequest(
            "You need to select at least one permission (read and/or write)."
        )

    try:
        workspace = workspace_api.get_by_id(str(workspace_id))
        for group in group_api.get_all_groups_by_list_id(groups_ids):
            if is_read_checked:
                workspace_api.add_group_read_access_to_workspace(
                    workspace, group, request.user
                )
            if is_write_checked:
                workspace_api.add_group_write_access_to_workspace(
                    workspace, group, request.user
                )
    except AccessControlError as ace:
        return HttpResponseBadRequest(escape(str(ace)))
    except DoesNotExist as dne:
        return HttpResponseBadRequest(escape(str(dne)))
    except Exception:
        return HttpResponseBadRequest("Something wrong happened.")

    return HttpResponse(json.dumps({}), content_type="application/javascript")


@method_decorator(login_required, name="dispatch")
class AssignView(View):
    """Assign Ajax view"""

    api = None

    def post(self, request):
        """Assign the record to a workspace.

        Args:
            request:

        Returns:
        """
        document_ids = request.POST.getlist("document_id[]", [])
        workspace_id = request.POST.get("workspace_id", None)

        if workspace_id is None or workspace_id == "":
            workspace = None
        else:
            try:
                workspace = workspace_api.get_by_id(str(workspace_id))
            except DoesNotExist:
                return HttpResponseBadRequest(
                    "The selected workspace does not exist anymore."
                )
            except Exception:
                return HttpResponseBadRequest("Something wrong happened.")

        for data_id in document_ids:
            try:
                self.api.assign(
                    self.api.get_by_id(data_id, request.user),
                    workspace,
                    request.user,
                )
            except AccessControlError as ace:
                return HttpResponseForbidden(escape(str(ace)))
            except Exception:
                return HttpResponseBadRequest("Something wrong happened.")

        return HttpResponse(
            json.dumps({}), content_type="application/javascript"
        )


def change_data_display(request):
    """Change data display

    Args:
        request:

    Returns:
    """
    try:
        xsl_transformation_id = request.POST.get("xslt_id", None)
        data_id = request.POST.get("data_id", None)
        if data_id:
            data = data_api.get_by_id(data_id, request.user)
            template_id = data.template.id
            template_hash = data.template.hash
            xml_content = data.xml_content
        else:
            template_id = request.POST.get("template_id", None)
            xml_content = request.POST.get("content", "")
            if template_id:
                template = template_api.get_by_id(template_id, request)
                template_hash = template.hash
            else:
                return HttpResponseBadRequest(
                    "missing parameters: template id is missing"
                )
        return HttpResponse(
            json.dumps(
                {
                    "template": render_xml_as_html_detail(
                        xml_content=xml_content,
                        template_id=template_id,
                        template_hash=template_hash,
                        xslt_id=xsl_transformation_id,
                        request=request,
                    ),
                }
            ),
            "application/javascript",
        )
    except AccessControlError:
        return HttpResponseForbidden("Access Forbidden")
    except Exception:
        return HttpResponseBadRequest("Unexpected error")


@method_decorator(login_required, name="dispatch")
class LoadBlobMetadataForm(View):
    """Load the form to add metadata files to a blob."""

    def get(self, request):
        """Load the form to add metadata files to a blob.

        Args:
            request:

        Returns:

        """
        blob_id = request.GET.get("blob_id", None)

        try:
            blob_object = blob_api.get_by_id(str(blob_id), request.user)
            form = BlobMetadataForm(user=request.user, blob=blob_object)
        except exceptions.ModelError:
            return HttpResponseBadRequest("Blob not found.")
        except Exception:
            return HttpResponseBadRequest("An unexpected error occurred.")

        context = {"add_metadata_form": form}

        return HttpResponse(
            json.dumps(
                {
                    "form": loader.render_to_string(
                        "core_main_app/user/blob/list/modals/add_metadata_form.html",
                        context,
                    )
                }
            ),
            "application/javascript",
        )


@method_decorator(login_required, name="dispatch")
class AddMetadataToBlob(View):
    """Add metadata files to a blob."""

    def post(self, request):
        """Add metadata files to a blob.

        Args:
            request:

        Returns:

        """
        blob_id = request.POST.get("blob_id", None)
        metadata_id_list = request.POST.getlist("metadata_id[]", [])

        try:
            blob_object = blob_api.get_by_id(str(blob_id), request.user)
            check_can_write(blob_object, request.user)
            blob_api.add_metadata_list(
                blob_object,
                [
                    data_api.get_by_id(metadata_id, request.user)
                    for metadata_id in metadata_id_list
                ],
                request.user,
            )
        except exceptions.ModelError:
            return HttpResponseBadRequest("Blob not found.")
        except AccessControlError:
            return HttpResponseBadRequest("Permission denied.")
        except Exception:
            return HttpResponseBadRequest("An unexpected error occurred.")

        messages.add_message(
            request,
            messages.SUCCESS,
            "Metadata updated.",
        )
        return HttpResponse(json.dumps({}), "application/javascript")


@method_decorator(login_required, name="dispatch")
class RemoveMetadataFromBlob(View):
    """Remove metadata file from blob."""

    def post(self, request):
        """Remove metadata file

        Args:
            request:

        Returns:

        """
        blob_id = request.POST.get("blob_id", None)
        metadata_id = request.POST.get("metadata_id", None)

        try:
            blob_object = blob_api.get_by_id(str(blob_id), request.user)
            check_can_write(blob_object, request.user)
            metadata = data_api.get_by_id(metadata_id, request.user)
            blob_api.remove_metadata(blob_object, metadata, request.user)
        except exceptions.ModelError:
            return HttpResponseBadRequest("Blob not found.")
        except AccessControlError:
            return HttpResponseBadRequest("Permission denied.")
        except Exception:
            return HttpResponseBadRequest("An unexpected error occurred.")

        messages.add_message(
            request,
            messages.INFO,
            "Metadata successfully removed.",
        )
        return HttpResponse(json.dumps({}), "application/javascript")


@method_decorator(login_required, name="dispatch")
class UploadFile(View):
    """Upload file"""

    def post(self, request):
        """Upload file

        Args:
            request:

        Returns:

        """
        try:
            form = BlobFileForm(request.POST, request.FILES)
            if form.is_valid():
                blob = Blob(
                    filename=form.files["file"].name,
                    blob=form.files["file"],
                    user_id=str(request.user.id),
                )
                blob_api.insert(blob, request.user)
                messages.add_message(
                    request,
                    messages.SUCCESS,
                    "File successfully uploaded.",
                )
                return HttpResponse(json.dumps({}))
            else:
                return HttpResponseBadRequest(
                    json.dumps({"message": str(form.errors.as_text())})
                )
        except Exception:
            messages.add_message(
                request,
                messages.ERROR,
                "An error occurred during file upload.",
            )

            return HttpResponseBadRequest(json.dumps({}))
